SystemOrganization addCategory: #'Ma special collections'!
SystemOrganization addCategory: #'Ma special collections-Info'!
SystemOrganization addCategory: #'Ma special collections-Integer collections'!
SystemOrganization addCategory: #'Ma special collections-Squeak weak tweaks'!
SystemOrganization addCategory: #'Ma special collections-Tree'!
SystemOrganization addCategory: #'Ma special collections-MaHashIndex'!
SystemOrganization addCategory: #'Ma special collections-private'!
SystemOrganization addCategory: #'Ma special collections-Tests'!
SystemOrganization addCategory: #'Ma special collections-Exceptions'!

MaTestCase subclass: #MaHashIndexRecordTester
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Tests'!

!MaHashIndexRecordTester methodsFor: 'tests' stamp: 'cmm 4/10/2005 22:55'!
testAll
	"Test a variety of keySize and recordSize combinations.  For each pair, run only as far as it takes to get two-deep into duplicates for all originating from one entry in the root record, chosen at random."

	| testCount |
	testCount _ 0.
	#(16 24 32 40 56 64 96 128 160 192 256) "keySizes" do:
		[ : keySize |
		#(2 4 10 30 50 75 100) "recordSizes" do:
			[ : recordSize |  | startRecord  |
			"make test run-time reasonable by skipping enough levels to only test the bottom 100000 records or so."
			startRecord _ 
				self
					zoomedRecord: keySize 
					recordSize: recordSize 
					childSpace: 10000.
			self notify: 'testing keySize: ', keySize printString, ' recordSize: ', recordSize printString, ', total key span: ', (startRecord highKey - startRecord lowKey) printString.
			self 
				verifyRecord: startRecord					
				do:
					[ : record : eachBax |
					record == startRecord
						ifTrue:
							[ self notify:
								(String streamContents:
									[ : stream |
									stream
										tab; maPrint: 'x:' ;
										maPrint: (record entryIndexForBax: eachBax) ;
										maPrint: ' (~' ;
										maPrint: testCount ;
										maPrint: ' records tested)' ]) ].
					testCount _ testCount + 1 ] ] ]! !

!MaHashIndexRecordTester methodsFor: 'private tests' stamp: 'cmm 3/11/2004 23:29'!
verifyConsecutiveContiguousKeyRangesOn: aMaHashIndexRecord

	0
		to: aMaHashIndexRecord numberOfEntries - 2
		do:
			[ :index |
			self should:
				[ aMaHashIndexRecord canHoldDuplicateKeys
					ifTrue:
						[
						((aMaHashIndexRecord lowestKeyAtIndex: index + 1) -
							(aMaHashIndexRecord highestKeyAtIndex: index))
								between: 0
								and: 1 ]
					ifFalse:
						[
						((aMaHashIndexRecord lowestKeyAtIndex: index + 1) -
							(aMaHashIndexRecord highestKeyAtIndex: index)) = 1 ] ] ]
! !

!MaHashIndexRecordTester methodsFor: 'private tests' stamp: 'cmm 3/11/2004 17:03'!
verifyEntryIndexAndRangeEndpointCongruenceOn: aMaHashIndexRecord

	0
		to: aMaHashIndexRecord numberOfEntries - 1
		do:
			[ : x |  | key |

			key _ aMaHashIndexRecord lowestKeyAtIndex: x.
			self should:
				[ x
					between: (aMaHashIndexRecord entryIndexFor: key)
					and: (aMaHashIndexRecord highestEntryIndexFor: key) ].

			x > 0 ifTrue:
				[ self should: [ (aMaHashIndexRecord highestEntryIndexFor: key-1) < x ] ].

			key _ aMaHashIndexRecord highestKeyAtIndex: x.
			self should:
				[ x
					between: (aMaHashIndexRecord entryIndexFor: key)
					and: (aMaHashIndexRecord highestEntryIndexFor: key) ].

			x < (aMaHashIndexRecord numberOfEntries - 1) ifTrue:
				[ self should: [ (aMaHashIndexRecord entryIndexFor: key+1) > x ] ] ]! !

!MaHashIndexRecordTester methodsFor: 'private tests' stamp: 'cmm 3/11/2004 23:34'!
verifyRecord: aMaHashIndexRecord do: twoArgBlock

	self
		verifyConsecutiveContiguousKeyRangesOn: aMaHashIndexRecord;
		verifyEntryIndexAndRangeEndpointCongruenceOn: aMaHashIndexRecord.
	aMaHashIndexRecord
		byteArrayIndexesFrom: 0
		do:
			[ : eachBax |
			twoArgBlock value: aMaHashIndexRecord value: eachBax.
			(aMaHashIndexRecord highKey - aMaHashIndexRecord lowKey) > 1  "if not fully zoomed in"
				ifTrue:
					[ | x lowestKey highestKey |
					x _ aMaHashIndexRecord entryIndexForBax: eachBax.
					lowestKey _ aMaHashIndexRecord lowestKeyAtIndex: x.
					highestKey _ aMaHashIndexRecord highestKeyAtIndex: x.
					self
						verifyRecord:
							(aMaHashIndexRecord independentCopy
								beZoomedInToward: (lowestKey to: highestKey) atRandom
								from: aMaHashIndexRecord)
						do: twoArgBlock ] ]! !

!MaHashIndexRecordTester methodsFor: 'private creation' stamp: 'cmm 4/28/2005 21:25'!
zoomedRecord: keySize recordSize: recordSize childSpace: anInteger

	| answer prior |
	answer _
		MaHashIndexRecord
			keySize: keySize
			valueSize: 64
			numberOfEntries: recordSize.
	[ (answer highKey - answer lowKey) > anInteger ]
		whileTrue:
			[ prior _ answer.
			answer _ answer
				beZoomedInToward: (answer lowKey to: answer highKey) atRandom
				from: answer ].
	^ (answer highKey - answer lowKey) < 1000
		ifTrue: [ prior ]
		ifFalse: [ answer ]! !

MaTestCase subclass: #MaHashIndexTester
	instanceVariableNames: 'sortedKeys predictedSize index values nextValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Tests'!

!MaHashIndexTester methodsFor: 'private actions' stamp: 'cmm 4/28/2005 22:08'!
addKey: keyInteger

	| value |
	self index
		add: (value _ self getNextValue)
		at: keyInteger.
	sortedKeys add: keyInteger.
	(values
		at: keyInteger
		ifAbsentPut: [ OrderedCollection new ])
			add: value.
	predictedSize _ predictedSize + 1.
	^ value! !

!MaHashIndexTester methodsFor: 'private actions' stamp: ''!
addKeys: anArray

	^anArray collect: [ :each | self addKey: each ]
! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 12/19/2002 21:25'!
addKeysToIndex: anInteger

	| r lowest highest big |

	r _ Random new.
	big _ ((2 raisedTo: 64) - 1).
	highest _ 0.
	lowest _ big.
	anInteger timesRepeat:
		[ | n |
		n _ (r next * big) asInteger.
		n < lowest ifTrue: [ lowest _ n ].
		n > highest ifTrue: [ highest _ n ].
		self addKey: n ]! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 4/28/2005 22:03'!
allValues

	^ values
		inject: OrderedCollection new
		into:
			[ : coll : each |
			coll
				addAll: each;
				yourself ]! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 4/28/2005 22:04'!
getNextValue

	^ nextValue _ nextValue + 1! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 1/30/2002 12:48'!
index

	^index
! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 4/11/2005 10:58'!
indexFileNameForKeySize: keySize recordSize: recordSize

	^ self environmentDirectory fullNameFor: 'k', keySize printString, 'r', recordSize printString, '.', MaHashIndex suggestedFileExtension! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 11/30/2002 17:41'!
indexKeys

	^(self index
		upTo: sortedKeys size
		keysFrom: 0) copy! !

!MaHashIndexTester methodsFor: 'private accessing' stamp: 'cmm 4/28/2005 22:05'!
indexValues

	^ (self index
		upTo: sortedKeys size
		keysAndValuesFrom: 0
		to: self index highestPossibleKey
		startingAt: 1) last copy! !

!MaHashIndexTester methodsFor: 'private initializing' stamp: 'cmm 4/28/2005 22:06'!
initialize

	super initialize.
	values _ Dictionary new.
	sortedKeys _ SortedCollection new.
	nextValue _ 2 raisedTo: 17  "some arbitrary number"! !

!MaHashIndexTester methodsFor: 'accessing' stamp: 'cmm 4/28/2005 22:30'!
keyAndRecordSizesToTest

	^#(32 40 56 64 96 128 160 192 256) -> #(2 4 10 30 50 75 100)! !

!MaHashIndexTester methodsFor: 'private actions' stamp: 'cmm 3/14/2004 13:33'!
keySizes: keySizesArray withRecordSizes: recordSizesArray run: selectorSymbol

	keySizesArray do:
		[ : eachKeySize |
		recordSizesArray do:
			[ : eachRecordSize |
			self setUp.
			index _ self
				newHashIndexWithKeySize: eachKeySize
				recordSize: eachRecordSize.
			self
				notify: (String streamContents:
					[ :stream |
					stream
						nextPutAll: 'keysize ';
						maPrint: eachKeySize;
						nextPutAll: ' record size ';
						maPrint: eachRecordSize;
						nextPutAll: ' completed in ';
						nextPutAll: [ self perform: selectorSymbol ] maCalendarTimeToRun printString ]);
			tearDown ] ]! !

!MaHashIndexTester methodsFor: 'private creation' stamp: 'cmm 4/28/2005 22:13'!
newHashIndexWithKeySize: keySize recordSize: recordSize

	| filename |
	self notify: 'creating ',
		(filename _ self
			indexFileNameForKeySize: keySize
			recordSize: recordSize), '..'.
	self deleteFilesNamed: { filename }.
	^ MaHashIndex
		create: filename
		keySize: keySize
		valueSize: 48
		recordSize: recordSize! !

!MaHashIndexTester methodsFor: 'private actions' stamp: ''!
removeKey: keyInteger

	^self
		removeKey: keyInteger
		at: 1
! !

!MaHashIndexTester methodsFor: 'private actions' stamp: 'cmm 4/28/2005 22:06'!
removeKey: keyInteger at: anInteger

	self index
		remove: ((values at: keyInteger) removeAt: anInteger)
		at: keyInteger.
	sortedKeys remove: keyInteger.
	(values at: keyInteger) isEmpty ifTrue: [ values removeKey: keyInteger ].
	predictedSize _ predictedSize - 1
! !

!MaHashIndexTester methodsFor: 'private actions' stamp: 'cmm 11/28/2002 00:15'!
removeKeys: aCollection

	aCollection do: [ :each | self removeKey: each ]! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 4/18/2005 09:57'!
runHeavyAddsAndRemoves

	| keys1 keys2 keysToRemove1 keysToRemove2 indexDef tempKeys1 tempKeys2 |
	keys1 _ OrderedCollection new: 40000.
	keys2 _ OrderedCollection new: 40000.
	indexDef _
		(MaAsciiStringIndexDefinition attribute: #maYourselfValue)
			keySize: index keySize ;
			yourself.
	Symbol allSymbolTablesDo:
		[ :e |
		2 atRandom even
			ifTrue: [ keys2 add: (indexDef indexHashFor: e) ]
			ifFalse: [ keys1 add: (indexDef indexHashFor: e) ] ].
	keysToRemove1 _ OrderedCollection new.
	tempKeys1 _ keys1 copy.  tempKeys2 _ keys2 copy.
	keys1 size // 3 timesRepeat: [ keysToRemove1 add: (tempKeys1 remove: tempKeys1 atRandom) ].
	keysToRemove2 _ OrderedCollection new.
	keys2 size // 3 timesRepeat: [ keysToRemove2 add: (tempKeys2 remove: tempKeys2 atRandom) ].
	self
		addKeys: keys1;
		removeKeys: keysToRemove1;
		addKeys: keys2;
		removeKeys: keysToRemove2;
		verifyIndex! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 4/18/2005 10:11'!
runInsertionOfEndpointKeys
	"This tests keys that fall at the endpoints of records.  It covers the lowest and highest keys."

	self
		verifyInsertionOfDuplicateKey: 0;
		verifyInsertionOfDuplicateKey: (2 raisedTo: index keySize) - 1.
	self verifyIndex! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 3/10/2004 01:09'!
runKeyInsertionIntoBroadRecords
	"A 'broad' record is one whose key range is greater than its
	numberOfEntries, thus it cannot hold duplicate keys.
	This test adds keys in an order that requires the index to
	re-sequence keys across multiple broad records."

	self
		addKeys: #(500 450 470 550 400);
		verifyIndex;
		removeKey: 450;
		verifyIndex;
		addKey: 471;
		verifyIndex.
	25 timesRepeat: [ self addKey: 470 ].
	self verifyIndex.
	25 timesRepeat: [ self addKey: 469 ].
	self verifyIndex.
	1
		to: 25
		do:
			[ :each |
			each even
				ifTrue:
					[ self
						removeKey: 469
						at: each // 2 ] ].
	self verifyIndex! !

!MaHashIndexTester methodsFor: 'private framework' stamp: 'cmm 4/28/2005 22:07'!
setUp

	MaHashIndex allInstances do: [ :each | each close ].
	predictedSize _ 0.
	values _ Dictionary new.
	sortedKeys _ SortedCollection new.
	nextValue _ 2 raisedTo: 17  "some arbitrary number"! !

!MaHashIndexTester methodsFor: 'private framework' stamp: 'cmm 4/16/2002 23:43'!
tearDown

	index ifNotNil: [ index close ]
! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 12/5/2002 17:28'!
testInsertionOfEndpointKeys

	self
		keySizes: self keyAndRecordSizesToTest key
		withRecordSizes: self keyAndRecordSizesToTest value
		run: #runInsertionOfEndpointKeys! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 12/5/2002 17:28'!
testKeyInsertionIntoBroadRecords

	self
		keySizes: self keyAndRecordSizesToTest key
		withRecordSizes: self keyAndRecordSizesToTest value
		run: #runKeyInsertionIntoBroadRecords! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 11/30/2002 19:19'!
verifyIndex

	self shouldnt: [ index hasAnomalies ].
	self
		verifySize;
		verifyKeys;
		verifyKeyValuePairs;
		verifyKeyQuantities;
		verifyIndexPositions! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 4/28/2005 22:03'!
verifyIndexPositions

	| allValues |

	allValues _ self allValues.
	(self index
		upTo: allValues size
		keysAndValuesFromIndex: 1) last
			maConsistsOf: allValues! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 4/17/2005 22:07'!
verifyInsertionOfDuplicateKey: keyInteger
	"Test the insertion of a many of the same key, requiring the
	receivers index to zoom all the way into that key and getting
	into the inverted records."

	25 timesRepeat: [ self addKey: keyInteger ].
	self should:
		[ (self index
			upTo: 100
			at: keyInteger) last
				size = 25 ].
	275 timesRepeat: [ self addKey: keyInteger ].
	self should:
		[ (self index
			upTo: 350
			at: keyInteger) last
				size = 300 ]
! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 4/28/2005 22:06'!
verifyKeyQuantities

	self should:
		[ sortedKeys allSatisfy: [ :each | (self index occurrencesOf: each) = (values at: each) size ] ].
	self should:
		[ self index size = (
			self index
				numberOfEntriesFrom: 0
				to: sortedKeys last) ]
! !

!MaHashIndexTester methodsFor: 'private tests' stamp: 'cmm 4/28/2005 22:06'!
verifyKeyValuePairs

	self indexKeys
		with: self indexValues
		do: [ : eachKey : eachValue | self should: [ (values at: eachKey) includes: eachValue ] ]
! !

!MaHashIndexTester methodsFor: 'private tests' stamp: ''!
verifyKeys

	self should: [ self indexKeys maMatchesElementsOf: sortedKeys ]
! !

!MaHashIndexTester methodsFor: 'private tests' stamp: ''!
verifySize

	^self should: [ self index size = predictedSize ]
! !

!MaHashIndexTester methodsFor: 'tests' stamp: 'cmm 4/18/2005 22:02'!
xtestHeavyAddsAndRemoves
"renamed this guy 'x' for now because it takes too long to run."
	self
		keySizes: self keyAndRecordSizesToTest key
		withRecordSizes: self keyAndRecordSizesToTest value
		run: #runHeavyAddsAndRemoves! !

MaUserError subclass: #MaHashIndexUserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Exceptions'!

PackageInfo subclass: #MaSpecialCollectionsPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Info'!

!MaSpecialCollectionsPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 1/29/2003 23:33'!
prerequisitePackages

	^#('Ma base additions')! !

!MaSpecialCollectionsPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 4/28/2005 22:59'!
readmeText

	^'I provide several independent, special-purpose collection classes that seem to be generally useful.

MaWeakIdentityKeyDictionary and MaWeakValueDictionary wrap a collection of their Smalltalk counterparts and keeps them all under 12-bits in size, thus minimizing the number of collisions due to Squeaks 12-bits identity hash.


MaTree is one of my first attempts at a Smalltalk collection.  I never actually used it though I decided to keep it around.  It''s API compatible with the other collections.  To play with an example:

	MaTree new
		add: ''Great Grandma'';
		add: ''Grandma'' asChildOf: ''Great Grandma'';
		add: ''Mom'' asChildOf: ''Grandma'';
		yourself


MaIntervalCollection is It can efficiently track interval ranges and merge them when they get within their proximityThreshold.  This is useful for recording occupied space in a large range or, perhaps grouping entities based on their proximity to each other.  Example:

	(MaIntervalCollection proximityThreshold: 1)
		addFrom: 5 to: 10;
		addFrom: 15 to: 20;
		addFrom: 25 to: 30;
		addFrom: 21 to: 24;
		yourself


MaLargeArrayOfNumbers is an auto-growing, never-shrinking array of numbers.  All numbers have the same number of bits available to represent them.  It maintains the collection in a single randomly-accessed file, and is therefore constrained by available disk space or the addressability of the file.

	MaLargeArrayOfNumbers create: filenameString bitSize: 32.
	myArray _ MaLargeArrayOfNumbers open: filenameString.
	myArray
		at: 10 put: 8734 ;
		at: 11 put: 6199 ;
		yourself
	"be sure to #close it when you are done."


MaHashIndex is another file-based, large-collection of uniformly-sized numbers.  It is much slower but more flexible than MaLargeArrayOfNumbers.  It allows pages between any two key ranges via:

	#upTo: countInteger keysAndValuesFrom: lowKey to: highKey startingAt: index

and pages between any two absolute positions via:

	#upTo: countInteger keysAndValuesFromIndex: lowIndex

Other features are:
	- supports key and value sizes from 32 to 256 bits
	- supports duplicate keys
	- includes test cases that demonstrate its capability and correctness

	MaHashIndex 
		create: fileNameString 
		keySize: 64 
		hashRecordSize: 10.
	myIndex _ MaHashIndex open: filenameString.
	myIndex
		add: 8734 at: 10 ;
		add: 1212 at: 10 ;  "duplicate key ok"
		add: 6199 at: 11 ;
		yourself.
	"be sure to #close it when you are done."
'! !

MaObject subclass: #MaDictionary
	instanceVariableNames: 'dicts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaDictionary commentStamp: 'cmm 12/19/2002 23:50' prior: 0!
I am identical to a Dictionary but a lot faster when my size grows beyond 4K elements.  Squeak uses a 12-bit identity hash for its object identity, causing lots of hash collisions once you get more than 4K (2 raisedTo: 12) objects in it.  As this script demonstrates, I am faster at adding, accessing and replacing.

| sd md r |
r _ (1 to: 5500) collect: [ :e | Object new ].
r _ r shuffled.
sd _ WeakIdentityKeyDictionary new.
md _ MaWeakIdentityKeyDictionary new.

Transcript cr; show: 'time to add to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.

Transcript cr; show: 'time to add to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.

r_r shuffled.

Transcript cr; show: 'time to access to sd: ', ([ r do: [ :each | sd at: each ] ] timeToRun) printString.

Transcript cr; show: 'time to access to md: ', ([ r do: [ :each | md at: each ] ] timeToRun) printString.

r _ r shuffled.

Transcript cr; show: 'time to replace to sd: ', ([ r do: [ :each | sd at: each put: each ] ] timeToRun) printString.

Transcript cr; show: 'time to replace to md: ', ([ r do: [ :each | md at: each put: each ] ] timeToRun) printString.
!

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/27/2002 10:09'!
new

	^self new: 8! !

!MaDictionary class methodsFor: 'creation' stamp: 'cmm 12/13/2002 14:28'!
new: anInteger

	^super new capacity: anInteger; yourself! !

!MaDictionary methodsFor: 'adding' stamp: 'cmm 12/12/2002 17:10'!
add: anAssociation

	self
		removeKey: anAssociation key
		ifAbsent: [ "do nothing" ].
	self qualifyingDict add: anAssociation! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/22/2002 18:26'!
associationsDo: aBlock

	dicts do: [ :each | each associationsDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:43'!
at: anObject

	^self
		at: anObject
		ifAbsent: [ self errorKeyNotFound ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:34'!
at: anObject ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| result |
		result _ each
			at: anObject
			ifAbsent: [ uniqueObject ].
		result == uniqueObject ifFalse: [ ^result ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:44'!
at: key ifAbsentPut: aBlock

	^self
		at: key
		ifAbsent:
			[ self 
				at: key 
				put: aBlock value ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 6/8/2004 12:51'!
at: key put: anObject

	key ifNil: [ self error: 'nil cannot be used as a key to a Dictionary.' ].
	dicts do:
		[ :each |
		(each includesKey: key) ifTrue: [ ^each at: key put: anObject ] ].
	^self qualifyingDict
		at: key
		put: anObject! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 09:57'!
capacity: anInteger

	| numberOfDicts |
	numberOfDicts _ anInteger // self criticalSize.
	dicts _ OrderedCollection new: numberOfDicts.
	numberOfDicts timesRepeat: [ dicts add: (self squeakDictionaryClass new: self criticalSize) ].
	anInteger \\ self criticalSize > 0
		ifTrue:
			[ dicts add: (self squeakDictionaryClass new: anInteger \\ self criticalSize) ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:17'!
collect: aBlock 
	"Evaluate aBlock with each of my values as the argument.  Collect the
	resulting values into a collection that is like me. Answer with the new
	collection."
	| newCollection |
	newCollection _ OrderedCollection new: self size.
	self do: [:each | newCollection add: (aBlock value: each)].
	^ newCollection! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:04'!
copy

	^ self copy postCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/27/2002 10:05'!
criticalSize
	"Since the Squeak Set allocates about 4/3rd's extra space for future growth, we have the potential to grow a little out of its efficiency range.  Compensate."

	^4096*3//4! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock

	^self
		detect: aBlock
		ifNone: [ self error: 'not in MaDictionary' ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 2/5/2003 20:47'!
detect: aBlock ifNone: notFoundBlock

	self do: [ :each | (aBlock value: each) ifTrue: [ ^each ] ].
	^notFoundBlock value! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:58'!
do: aBlock

	dicts do: [ :each | each do: aBlock ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:43'!
errorKeyNotFound

	self error: 'key not found'! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/13/2002 14:03'!
finishCopy

	dicts _ dicts collect: [ :each | each copy ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includes: anObject

	self do: [:each | anObject = each ifTrue: [^true]].
	^false! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:14'!
includesKey: key 
	"Answer whether the receiver has a key equal to the argument, key."
	
	self at: key ifAbsent: [^false].
	^true! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 16:52'!
keys
	"Answer a Set containing the receiver's keys."
	| aSet |
	aSet _ Set new: self size.
	self keysDo: [:key | aSet add: key].
	^ aSet! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/25/2002 23:28'!
keysAndValuesDo: twoArgBlock

	dicts do: [ :each | each keysAndValuesDo: twoArgBlock ]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:06'!
keysAndValuesRemove: twoArgBlock

	dicts do: [ :each | each keysAndValuesRemove: twoArgBlock ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:52'!
keysDo: aBlock

	dicts do: [ :each | each keysDo: aBlock ]! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:12'!
maAllKeysAtValue: anObject

	| aCollection |
	aCollection := OrderedCollection new.
	self keysAndValuesDo: [ :key :value | value = anObject ifTrue: [ aCollection add: key ] ].
	^aCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/13/2002 14:33'!
maAt: anObject ifPresent: oneArgBlock ifAbsent: zeroArgBlock

	| current uniqueObject |

	uniqueObject _ ''.
	^(current := self
		at: anObject
		ifAbsent: [ uniqueObject ]) == uniqueObject
			ifTrue: [ zeroArgBlock value ]
			ifFalse: [ oneArgBlock value: current ]! !

!MaDictionary methodsFor: 'testing' stamp: 'cmm 12/12/2002 17:15'!
maIsHashedCollection

	^true! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:16'!
occurrencesOf: anObject 
	"Answer how many of the receiver's elements are equal to anObject."

	| count |
	count _ 0.
	self do: [:each | anObject = each ifTrue: [count _ count + 1]].
	^count! !

!MaDictionary methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:04'!
postCopy

	dicts _ dicts copy.
	self finishCopy! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 23:20'!
qualifyingDict

	^dicts
		detect: [ :each | each size < self criticalSize ]
		ifNone: [ dicts addFirst: self squeakDictionaryClass new ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 17:15'!
rehash

	dicts do: [ :each | each rehash ]! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 1/9/2003 00:40'!
reject: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifFalse: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject

	self shouldNotImplement! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:04'!
remove: anObject ifAbsent: exceptionBlock

	self shouldNotImplement! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/12/2002 17:05'!
removeKey: key 
	"Remove key from the receiver.
	If key is not in the receiver, notify an error."

	^ self removeKey: key ifAbsent: [self errorKeyNotFound]! !

!MaDictionary methodsFor: 'removing' stamp: 'cmm 12/13/2002 14:34'!
removeKey: key ifAbsent: aBlock

	| uniqueObject |
	uniqueObject _ ''.
	dicts do:
		[ :each |
		| removed |
		removed _ each
			removeKey: key
			ifAbsent: [ uniqueObject ].
		removed == uniqueObject ifFalse: [ ^removed ] ].
	^aBlock value! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 17:09'!
select: aBlock 
	"Evaluate aBlock with each of my values as the argument. Collect into a
	new dictionary, only those associations for which aBlock evaluates to
	true."

	| newCollection |
	newCollection _ self species new.
	self associationsDo: 
		[:each | 
		(aBlock value: each value) ifTrue: [newCollection add: each]].
	^newCollection! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 22:38'!
size

	^dicts
		inject: 0
		into: [ :sum :each | sum + each size ]! !

!MaDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 16:48'!
squeakDictionaryClass

	^Dictionary! !

!MaDictionary methodsFor: 'accessing' stamp: 'cmm 12/12/2002 17:11'!
values
	"Answer a Collection containing the receiver's values."
	| out |
	out _ WriteStream on: (Array new: self size).
	self valuesDo: [:value | out nextPut: value].
	^ out contents! !

!MaDictionary methodsFor: 'enumerating' stamp: 'cmm 12/12/2002 16:57'!
valuesDo: aBlock

	dicts do: [ :each | each valuesDo: aBlock ]! !

MaDictionary subclass: #MaWeakIdentityKeyDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 1/12/2003 21:26'!
finalizeValues

	dicts do: [ :e | e finalizeValues ]! !

!MaWeakIdentityKeyDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakIdentityKeyDictionary! !

MaDictionary subclass: #MaWeakValueDictionary
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Squeak weak tweaks'!

!MaWeakValueDictionary methodsFor: 'private' stamp: 'cmm 12/12/2002 22:27'!
squeakDictionaryClass

	^WeakValueDictionary! !

MaObject subclass: #MaHashIndexRecord
	instanceVariableNames: 'recordNumber lowKey highKey keySize valueSize byteArray numberOfEntries entrySize entryWeight canHoldDuplicateKeys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-private'!

!MaHashIndexRecord commentStamp: 'cmm 4/28/2005 21:25' prior: 0!
keySize and valueSize are specified in bits.
entrySize is in bytes.
!

!MaHashIndexRecord class methodsFor: 'creation' stamp: 'cmm 4/28/2005 21:23'!
keySize: keyBits valueSize: valueBits numberOfEntries: entrySizeInteger

	^ self new
		keySize: keyBits
		valueSize: valueBits
		numberOfEntries: entrySizeInteger ;
		postInitialize! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 12/6/2002 00:05'!
addAnomaliesTo: aDictionary usingParent: parentRecord

	| originalCount |
	originalCount _ aDictionary size.
	self
		addChildCountAnomaliesTo: aDictionary
		using: parentRecord.
	^aDictionary size > originalCount! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 12/6/2002 00:04'!
addChildCountAnomaliesTo: aDictionary using: parentRecord

	| bax parentNumberOfChildren isCorrect |
	bax _ self baxInParent: parentRecord for: self.
	parentNumberOfChildren _ (parentRecord numberOfChildrenAtByteArrayIndex: bax).
	isCorrect _ (self lowKey = self highKey or: [ self numberOfChildren = 0 ])
		ifTrue: 
			[ "it's possible for the record not to have an entry in the parent under this condition."
			(parentRecord hasEntryAtByteArrayIndex: bax)
				ifTrue: [ (parentNumberOfChildren - 1) = self numberOfChildren ]
				ifFalse: [ parentNumberOfChildren = self numberOfChildren ] ]
		ifFalse:
			[ (parentNumberOfChildren - 1) = self numberOfChildren ].
	isCorrect ifFalse:
		[ (aDictionary
			at: 'number of children discrepency'
			ifAbsentPut: [ OrderedCollection new ]) add: { parentRecord.  self } ]! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
allNumberOfChildren
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self numberOfChildrenAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 12/5/2002 23:52'!
anomaliesUsingParent: parentRecord

	| answer |
	answer _ OrderedCollection new.
	self
		addChildCountAnomaliesTo: answer
		using: parentRecord.
	^answer! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/10/2005 21:21'!
approximateEntryIndexFor: keyInteger
	"Answers the entryIndex for keyInteger that is close but, due to rounding/truncation, may be off by one."

	^ (keyInteger - lowKey) * numberOfEntries // entryWeight! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 4/28/2005 21:56'!
atKey: keyInteger put: valueInteger
	"The receiver has been verified for some available slot for keyInteger."

	self
		eligibleByteArrayIndexesFor: keyInteger
		reverseDo:
			[ : index |
			(self hasSlotAvailableAtByteArrayIndex: index)
				ifTrue:
					[ self
						keyAtByteArrayIndex: index
						put: keyInteger.
					self
						valueAtByteArrayIndex: index
						put: valueInteger.
					^ index ] ]! !

!MaHashIndexRecord methodsFor: 'SUnit tests' stamp: 'cmm 11/29/2002 23:39'!
baxInParent: parentRecord for: childRecord

	parentRecord
		byteArrayIndexesFrom: 0
		do:
			[ :bax |
			(parentRecord recordNumberAtByteArrayIndex: bax) = childRecord recordNumber
				ifTrue: [ ^bax ] ].
	self error: 'I am not child of parentRecord'! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: ''!
baxMap

	| map |

	map _ OrderedCollection new.
	self
		byteArrayIndexesFrom: 0
		do: [ :each | map add: (self entryIndexForBax: each) -> each ].
	^map asArray
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 3/30/2005 10:25'!
beFrom: lowKeyInteger to: highKeyInteger

	self
		lowKey: lowKeyInteger;
		highKey: highKeyInteger.
	self setLowKey: lowKeyInteger highKey: highKeyInteger! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
beZoomedInToward: keyInteger from: aMaHashIndexRecord
	"Set the receivers low and highIndex to the correct range that can
	accomodate the index where keyInteger belongs in aMaHashIndexRecord."

	| index |

	index _ aMaHashIndexRecord entryIndexFor: keyInteger.
	self
		beFrom: (aMaHashIndexRecord lowestKeyAtIndex: index)
		to: (aMaHashIndexRecord highestKeyAtIndex: index)
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
bufferSize

	^byteArray size
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
byteArray

	^byteArray! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/22/2002 00:12'!
byteArrayIndexContainingGlobalIndex: anInteger base: baseInteger
	"baseInteger is 0-based."

	| count |

	count _ baseInteger.
	self
		byteArrayIndexesFrom: 0
		do:
			[ :each |
			count _ count + (self numberOfChildrenAtByteArrayIndex: each).
			count >= anInteger ifTrue: [ ^each ] ].
	^nil! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
byteArrayIndexFor: entryIndex

	^entryIndex * entrySize + self headerSize
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/27/2002 14:55'!
byteArrayIndexOfLowestEntry
	"Answers nil if no entries on this record."

	self
		byteArrayIndexesFrom: 0
		do:
			[ :each | (self hasEntryAtByteArrayIndex: each) ifTrue: [ ^each ] ].
	^nil! !

!MaHashIndexRecord methodsFor: 'actions' stamp: ''!
byteArrayIndexesFrom: startingEntryIndex do: oneArgBlock

	self
		byteArrayIndexesFrom: startingEntryIndex
		to: numberOfEntries - 1
		do: oneArgBlock
! !

!MaHashIndexRecord methodsFor: 'actions' stamp: ''!
byteArrayIndexesFrom: startingEntryIndex to: endingEntryIndex do: oneArgBlock

	(self byteArrayIndexFor: startingEntryIndex)
		to: (self byteArrayIndexFor: endingEntryIndex)
		by: entrySize
		do: oneArgBlock
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
canAdd: keyInteger

	self
		eligibleByteArrayIndexesFor: keyInteger
		do: [ :index | (self hasSlotAvailableAtByteArrayIndex: index) ifTrue: [ ^true ] ].
	^false
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
canHoldDuplicateKeys

	^canHoldDuplicateKeys! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
clearAllEntries

	self
		byteArrayIndexesFrom: 0
		do: [ :each | self removeEntryAtByteArrayIndex: each ]
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 12/23/2002 14:42'!
decrementNumberOfChildrenAtByteArrayIndex: anInteger

	| currentNumberOfChildren |

	currentNumberOfChildren _ self numberOfChildrenAtByteArrayIndex: anInteger.
	self
		numberOfChildrenAtByteArrayIndex: anInteger
		put: currentNumberOfChildren - 1! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 11/30/2002 23:51'!
decrementNumberOfChildrenAtKey: keyInteger

	self decrementNumberOfChildrenAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 3/9/2004 23:53'!
dumpEntries

	self dumpEntriesTo: Transcript! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 4/28/2005 21:52'!
dumpEntriesTo: aStream

	self 
		byteArrayIndexesFrom: 0
		do:
			[ : eachBax |
			(self numberOfChildrenAtByteArrayIndex: eachBax) > 0
				ifTrue:
					[ | x |
					x _ self entryIndexForBax: eachBax.
					aStream cr; space; space;
						maPrint: 'x:'; maPrint: x;
						maPrint: ' ['; maPrint: (self lowestKeyAtIndex: x); maPrint: '..'; maPrint: (self highestKeyAtIndex: x); maPrint: ']' ;
						maPrint: ' bx:'; maPrint: (self byteArrayIndexFor: x) ;
						maPrint: ' k:'; maPrint: (self keyAtByteArrayIndex: eachBax);
						maPrint: ' o:'; maPrint: (self valueAtByteArrayIndex: eachBax);
						maPrint: ' r:'; maPrint: (self recordNumberAtByteArrayIndex: eachBax);
						maPrint: ' c:'; maPrint: (self numberOfChildrenAtByteArrayIndex: 
eachBax) ] ]! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 4/28/2005 21:47'!
dumpStructureTo: aStream
	self
		maPrintAttributes: #(
			recordNumber lowKey highKey canHoldDuplicateKeys keys recordNumbers values
			allNumberOfChildren possibleKeys)
		on: aStream! !

!MaHashIndexRecord methodsFor: 'actions' stamp: ''!
eligibleByteArrayIndexesFor: keyInteger do: oneArgBlock

	(self lowestByteArrayIndexFor: keyInteger)
		to: (self highestByteArrayIndexFor: keyInteger)
		by: entrySize
		do: [ :index | oneArgBlock value: index ]
! !

!MaHashIndexRecord methodsFor: 'actions' stamp: 'cmm 3/10/2004 01:00'!
eligibleByteArrayIndexesFor: keyInteger reverseDo: oneArgBlock

	(self highestByteArrayIndexFor: keyInteger)
		to: (self lowestByteArrayIndexFor: keyInteger)
		by: 0-entrySize
		do: [ :index | oneArgBlock value: index ]
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/12/2004 12:18'!
entryIndexFor: keyInteger
	"The lowest entry index in which keyInteger can reside."

	| approximate |
	approximate _ self approximateEntryIndexFor: keyInteger.
	^ keyInteger > (self highestKeyAtIndex: approximate)
		ifTrue: [ approximate+1 ]
		ifFalse: [ approximate ]! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
entryIndexForBax: anInteger
	"Bax means 'byteArrayIndex'."

	^(anInteger - self headerSize) / entrySize
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 3/12/2004 15:37'!
entryWeight

	^ entryWeight! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasChildAtByteArrayIndex: byteArrayIndex

	^(self recordNumberAtByteArrayIndex: byteArrayIndex) > 0
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: 'cmm 12/1/2002 21:44'!
hasChildAtHighestByteArrayIndexForBax: byteArrayIndex

	^(self recordNumberAtHighestByteArrayIndexForBax: byteArrayIndex) > 0! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasEntryAtByteArrayIndex: anInteger

	^(self hasSlotAvailableAtByteArrayIndex: anInteger) not
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: ''!
hasEntryForKey: keyInteger
	"Answer whether keyInteger resides in an entry of the receiver."

	self
		eligibleByteArrayIndexesFor: keyInteger
		do:
			[ :index |
			((self hasEntryAtByteArrayIndex: index)
				and: [ (self keyAtByteArrayIndex: index) = keyInteger ])
					ifTrue:
						[ ^true ] ].
	^false
! !

!MaHashIndexRecord methodsFor: 'testing' stamp: 'cmm 4/28/2005 21:52'!
hasSlotAvailableAtByteArrayIndex: anInteger

	^(self valueAtByteArrayIndex: anInteger) = self specialValueForUnusedEntry
! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
headerSize
	"Low and high keys.
	^(keySize / 8) * 2  simplified to / 4 for max performance."

	^keySize / 4
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
highKey

	^byteArray
		maUint: keySize
		at: keySize / 8
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
highKey: anInteger

	^byteArray
		maUint: keySize
		at: (keySize / 8)
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
highestByteArrayIndexFor: keyInteger
	"This will vary from the lowestByteArrayIndexFor: keyInteger only when 
	the receiver canHoldDuplicateKeys."

	^self byteArrayIndexFor: (self highestEntryIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 12/1/2002 21:18'!
highestByteArrayIndexForAnyKeyAt: byteArrayIndex
	"this is for records that canHoldDuplicateKeys when we need to find the next child record number, since that is stored at the highestByteArrayIndexFor: that key."

	^self highestByteArrayIndexFor: (self lowestKeyAtIndex: (self entryIndexForBax: byteArrayIndex))! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 3/12/2004 12:50'!
highestEntryIndexFor: keyInteger

	^canHoldDuplicateKeys
		ifTrue:
			[ (self entryIndexFor: keyInteger + 1) - 1 ]
		ifFalse:
			[ self entryIndexFor: keyInteger ]! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/12/2004 10:51'!
highestKeyAtIndex: anInteger

	^canHoldDuplicateKeys
		ifTrue:
			[ self lowestKeyAtIndex: anInteger ]
		ifFalse:
			[ (self lowestKeyAtIndex: anInteger + 1) - 1 ]
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/11/2004 21:19'!
highestPossibleKey

	^ (2 raisedTo: keySize) - 1! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
incrementNumberOfChildrenAtByteArrayIndex: anInteger

	| currentNumberOfChildren |

	currentNumberOfChildren _ self numberOfChildrenAtByteArrayIndex: anInteger.
	self
		numberOfChildrenAtByteArrayIndex: anInteger
		put: currentNumberOfChildren + 1
! !

!MaHashIndexRecord methodsFor: 'copying' stamp: 'cmm 3/24/2005 12:03'!
independentCopy

	^ self copy postIndependentCopy! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
keyAtByteArrayIndex: anInteger

	^byteArray
		maUint: keySize
		at: anInteger + self keyEntryOffset
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
keyAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: keySize
		at: index + self keyEntryOffset
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
keyAtEntryIndex: anInteger

	^self keyAtByteArrayIndex: (self byteArrayIndexFor: anInteger)
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
keyAtHighestSlotFor: keyInteger

	^self keyAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 22:44'!
keyEntryOffset

	^ self recordNumberEntryOffset + self sizeOfRecordNumber! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 3/11/2004 21:53'!
keySize

	^ keySize! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 4/28/2005 22:40'!
keySize: keyBits valueSize: valueBits numberOfEntries: numberOfEntriesInteger

	keySize _ keyBits.
	valueSize _ valueBits.
	numberOfEntries _ numberOfEntriesInteger.
	"Each entry consists of:  value (at: 0), record number (at: 8), key (at: 12), number of child entries (at: 20)."
	entrySize _ (valueSize/8) + self sizeOfRecordNumber + ((keySize / 8) * 2).
	byteArray _ ByteArray new: self headerSize + (entrySize * numberOfEntriesInteger).
	self clearAllEntries! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
keys
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self keyAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
lowKey

	^byteArray
		maUint: keySize
		at: 0
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 1/27/2003 23:57'!
lowKey: anInteger

	^byteArray
		maUint: keySize
		at: 0
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'private' stamp: ''!
lowestByteArrayIndexFor: keyInteger

	^self byteArrayIndexFor: (self entryIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/10/2005 23:14'!
lowestKeyAtIndex: entryIndex
	"Answer the lowest-possible key that can be stored at entryIndex."

	^ lowKey + (entryIndex * entryWeight // numberOfEntries)! !

!MaHashIndexRecord methodsFor: 'printing' stamp: 'cmm 3/12/2004 11:34'!
maPrintAbbreviatedOn: aStream

	self
		maPrintAttributes: #(lowKey highKey entryWeight possibleKeys canHoldDuplicateKeys)
		on: aStream! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
numberOfChildren

	| total |

	total _ 0.
	self
		byteArrayIndexesFrom: 0
		do: [ :index | total _ total + (self numberOfChildrenAtByteArrayIndex: index) ].
	^total
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:51'!
numberOfChildrenAtByteArrayIndex: anInteger

	^byteArray
		maUint: keySize
		at: anInteger + self numberOfChildrenEntryOffset
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:57'!
numberOfChildrenAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: keySize
		at: index + self numberOfChildrenEntryOffset
		put: anInteger! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 12/6/2002 15:45'!
numberOfChildrenBefore: keyInteger

	| lastByteArrayIndex entryIndex totalInSlotsBeforeKey |

	entryIndex _ self entryIndexFor: keyInteger.
	lastByteArrayIndex _ self byteArrayIndexFor: entryIndex.
	totalInSlotsBeforeKey _ self numberOfChildrenBeforeByteArrayIndex: lastByteArrayIndex.
	^((self keyAtByteArrayIndex: lastByteArrayIndex) < keyInteger
		and: [ self hasEntryAtByteArrayIndex: lastByteArrayIndex ])
			ifTrue:
				[ (lastByteArrayIndex = (self highestByteArrayIndexFor: keyInteger))
					ifTrue: [ totalInSlotsBeforeKey + 1 ]
					ifFalse: [ totalInSlotsBeforeKey + (self numberOfChildrenAtByteArrayIndex: lastByteArrayIndex) ] ]
			ifFalse: [ totalInSlotsBeforeKey ]! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 11/22/2002 00:13'!
numberOfChildrenBeforeByteArrayIndex: anInteger

	| total |

	total _ 0.
	self
		byteArrayIndexesFrom: 0
		do:
			[ :each |
			each = anInteger ifTrue: [ ^total ].
			total _ total + (self numberOfChildrenAtByteArrayIndex: each) ].
	self error: anInteger printString, ' is not a valid byteArrayIndex.'! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 12/30/2002 21:20'!
numberOfChildrenEntryOffset

	^self keyEntryOffset + (keySize / 8)! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
numberOfEntries

	^numberOfEntries
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 3/14/2004 13:43'!
populateFrom: aMaHashIndexRecord

	byteArray
		replaceFrom: 1
		to: byteArray size
		with: aMaHashIndexRecord byteArray.
	recordNumber _ aMaHashIndexRecord recordNumber.
	self 
		setLowKey: aMaHashIndexRecord primLowKey 
		highKey: aMaHashIndexRecord primHighKey! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: ''!
possibleKeys

	| answer |

	answer _ WriteStream on: (String new: numberOfEntries * 2).
	0
		to: numberOfEntries - 1
		do:
			[ :index | | low high |
			low _ self lowestKeyAtIndex: index.
			high _ self highestKeyAtIndex: index.
			answer nextPutAll: ' | '.
			low = high
				ifTrue:
					[ answer maPrint: low ]
				ifFalse:
					[ answer
						maPrint: low;
						maPrint: '..';
						maPrint: high ] ].
	^answer contents
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 3/24/2005 12:03'!
postIndependentCopy

	byteArray _ byteArray copy! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 3/11/2004 21:17'!
postInitialize

	self
		beFrom: 0
		to: self highestPossibleKey! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
primHighKey

	^highKey
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
primLowKey

	^lowKey
! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 3/14/2004 13:43'!
readFrom: file atPosition: anInteger

	file
		maRead: self bufferSize
		bytesFromPosition: 1
		of: byteArray
		atFilePosition: anInteger.
	self 
		setLowKey: self lowKey
		highKey: self highKey! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
recordNumber

	^recordNumber! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
recordNumber: anInteger

	recordNumber _ anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 1/27/2003 23:51'!
recordNumberAtByteArrayIndex: anInteger

	^byteArray 
		maUint: 32
		at: anInteger + self recordNumberEntryOffset
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 1/27/2003 23:58'!
recordNumberAtByteArrayIndex: index put: anInteger

	^byteArray
		maUint: 32
		at: index + self recordNumberEntryOffset
		put: anInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: 'cmm 12/1/2002 20:57'!
recordNumberAtHighestByteArrayIndexForBax: anInteger

	^self recordNumberAtByteArrayIndex: (self highestByteArrayIndexForAnyKeyAt: anInteger)! !

!MaHashIndexRecord methodsFor: 'building' stamp: ''!
recordNumberAtSlotForKey: keyInteger put: recordNumberInteger

	self
		recordNumberAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
		put: recordNumberInteger
! !

!MaHashIndexRecord methodsFor: 'accessing' stamp: ''!
recordNumberCloserTo: keyInteger

	^self recordNumberAtByteArrayIndex: (self highestByteArrayIndexFor: keyInteger)
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 22:41'!
recordNumberEntryOffset

	^ valueSize/8! !

!MaHashIndexRecord methodsFor: 'printing' stamp: ''!
recordNumbers
	"This is intended to be used only for printing!!"

	^(self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ :index | self recordNumberAtByteArrayIndex: index ]
! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:51'!
removeEntryAtByteArrayIndex: anInteger

	| answer |

	answer _ self valueAtByteArrayIndex: anInteger.
	self
		valueAtByteArrayIndex: anInteger
		put: self specialValueForUnusedEntry.
	^ answer! !

!MaHashIndexRecord methodsFor: 'building' stamp: 'cmm 4/28/2005 21:48'!
removeKey: keyInteger withValue: valueInteger

	self
		eligibleByteArrayIndexesFor: keyInteger
		do:
			[ : each |
			((self keyAtByteArrayIndex: each) = keyInteger
				and: [ (self valueAtByteArrayIndex: each) = valueInteger ])
					ifTrue:
						[ self
							removeEntryAtByteArrayIndex: each;
							decrementNumberOfChildrenAtByteArrayIndex: each.
						^ each ] ].
	^ nil! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/10/2005 21:20'!
setLowKey: lowKeyInteger highKey: highKeyInteger

	lowKey _ self lowKey.
	highKey _ self highKey.
	canHoldDuplicateKeys _ (highKey - lowKey + 1) < numberOfEntries.
	entryWeight _ highKey - lowKey + 1! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 12/30/2002 21:35'!
sizeOfRecordNumber

	^4! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:51'!
specialValueForUnusedEntry

	^ 65539! !

!MaHashIndexRecord methodsFor: 'debugging' stamp: 'cmm 3/11/2004 21:41'!
structure

	^ String streamContents: [ : stream | self dumpStructureTo: stream ]! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:44'!
valueAtByteArrayIndex: anInteger

	^ byteArray 
		maUint: valueSize
		at: anInteger + self valueEntryOffset! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:44'!
valueAtByteArrayIndex: index put: anInteger

	^ byteArray
		maUint: valueSize
		at: index + self valueEntryOffset
		put: anInteger! !

!MaHashIndexRecord methodsFor: 'private' stamp: 'cmm 4/28/2005 21:44'!
valueEntryOffset

	^ 0
! !

!MaHashIndexRecord methodsFor: 'printing' stamp: 'cmm 4/28/2005 21:52'!
values
	"This is intended to be used only for printing!!"

	^ (self headerSize
		to: byteArray size - 1
		by: entrySize)
			collect: [ : index | self valueAtByteArrayIndex: index ]! !

!MaHashIndexRecord methodsFor: 'actions' stamp: 'cmm 1/27/2002 22:46'!
writeTo: file atPosition: anInteger

	file
		maWrite: self bufferSize
		bytesFromPosition: 1
		of: byteArray
		atFilePosition: anInteger
! !

MaObject subclass: #MaHashIndexRecordFactory
	instanceVariableNames: 'index recordNumbersPath entryIndexPath byteArrayIndexPath recordPath preMadeNewRecord'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-private'!

!MaHashIndexRecordFactory class methodsFor: 'creation' stamp: 'cmm 12/4/2002 23:31'!
index: aMaHashIndex

	^self new
		index: aMaHashIndex;
		postInitialize;
		yourself! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 4/28/2005 21:24'!
createPreMadeNewRecord

	preMadeNewRecord :=
		MaHashIndexRecord
			keySize: index keySize
			valueSize: index valueSize
			numberOfEntries: index hashRecordSize
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
entryIndexAtLevel: anInteger

	^entryIndexPath at: anInteger
! !

!MaHashIndexRecordFactory methodsFor: 'private' stamp: 'cmm 12/4/2002 23:35'!
index: aMaHashIndex

	index := aMaHashIndex! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 4/28/2005 21:24'!
initializeRecordPathRecords

	1
		to: self maxDepth
		do:
			[ :x |
			recordPath
				at: x
				put: (
					MaHashIndexRecord
						keySize: index keySize
						valueSize: index valueSize
						numberOfEntries: index hashRecordSize) ]! !

!MaHashIndexRecordFactory methodsFor: 'private' stamp: 'cmm 12/31/2002 14:55'!
maxDepth

	^(index highestPossibleKey log: index numberOfEntries) asInteger! !

!MaHashIndexRecordFactory methodsFor: 'initializing' stamp: 'cmm 12/4/2002 23:35'!
postInitialize

	recordNumbersPath := OrderedCollection new: self maxDepth.
	entryIndexPath := OrderedCollection new: self maxDepth.
	byteArrayIndexPath := OrderedCollection new: self maxDepth.
	recordPath := Array new: self maxDepth.
	self
		createPreMadeNewRecord;
		initializeRecordPathRecords! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
preMadeNewRecord

	^preMadeNewRecord! !

!MaHashIndexRecordFactory methodsFor: 'building' stamp: ''!
put: aMaHashIndexRecord atLevel: levelInteger byteArrayIndex: indexInteger

	levelInteger > recordNumbersPath size
		ifTrue:
			[ recordNumbersPath add: aMaHashIndexRecord recordNumber.
			byteArrayIndexPath add: indexInteger.
			entryIndexPath add: (aMaHashIndexRecord entryIndexForBax: indexInteger) ]
		ifFalse:
			[ recordNumbersPath
				at: levelInteger
				put: aMaHashIndexRecord recordNumber.
			byteArrayIndexPath
				at: levelInteger
				put: indexInteger.
			entryIndexPath
				at: levelInteger
				put: (aMaHashIndexRecord entryIndexForBax: indexInteger) ].
	levelInteger > recordPath size
		ifFalse:
			[ (recordPath at: levelInteger) populateFrom: aMaHashIndexRecord ]
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: 'cmm 11/29/2002 22:47'!
recordAtLevel: levelInteger

	^levelInteger > recordPath size
		ifTrue:
			[ recordPath first independentCopy ]
		ifFalse:
			[ recordPath at: levelInteger ]
! !

!MaHashIndexRecordFactory methodsFor: 'accessing' stamp: ''!
recordNumberAtLevel: anInteger

	^recordNumbersPath at: anInteger
! !

MaObject subclass: #MaIntervalCollection
	instanceVariableNames: 'intervals proximityThreshold'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Integer collections'!

!MaIntervalCollection commentStamp: '<historical>' prior: 0!
This class keeps track of a collection of intervals.

As intervals are added that intersect or are within the proximityThreshold
of existing intervals, the existing interval is enlarged to include for the new interval.

If an interval is removed in the middle of a large interval such that
a "hole" is created, a new interval will be created and the other one
will be adjusted so that they will represent only the valid ranges.

For example:

	if an instance of this class has 1..10

	and we add: 8..15, then this class will have a single interval from 1..15

	Then remove 5..11..  Assuming a proximityThreshold of 1, this class will have the following two intervals:

		1..4
		12..15

Use a proximityThreshold of 0 to track the real-numbers..!

!MaIntervalCollection class methodsFor: 'creation' stamp: ''!
proximityThreshold: aMagnitude

	^self new proximityThreshold: aMagnitude! !

!MaIntervalCollection methodsFor: 'building' stamp: ''!
add: anInterval

	self
		addFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 8/18/2002 00:59'!
addFrom: from to: to

	| physicalFromIndex physicalToIndex |

	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	(from = to
		and: [ self isStartOfGap: physicalFromIndex ])
			ifTrue:
				[ "User does not realize that this class deals with real-number-based
				intervals that are inclusive on the left and exclusive on the right.
				Take care of him by adding a sensible interval."
				^self addFrom: from to: to + proximityThreshold ].
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "if in a gap, add a new interval"
			((self isStartOfGap: physicalFromIndex)
				and: [ (self
					fromValue: from
					isOutsideProximityThresholdAt: physicalFromIndex)
				and:
					[ self
						toValue: to
						isOutsideProximityThresholdAt: physicalToIndex ] ])
				ifTrue:
					[ self
						privateInsertFrom: from
						to: to
						afterIndex: physicalFromIndex.
					^from ] ].
	self
		extendIntervalFrom: from
		at: physicalFromIndex
		to: to
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 2/20/2002 01:17'!
bridgeFrom: physicalBridgeFromIndex to: physicalBridgeToIndex
	"Remove intervals in the receiver that are between physicalFromIndex
	and physicalToIndex which should be merged into one larger interval.
	The caller must ensure the from is the start of an interval and the to is
	the start of a gap."

	physicalBridgeToIndex - physicalBridgeFromIndex - 1
		timesRepeat: [ intervals removeAt: physicalBridgeFromIndex + 1 ]! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 12/20/2002 09:59'!
clear

	self
		removeFrom: self lowest
		to: self highest! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 9/24/2002 12:30'!
extendGapFromValue: from at: physicalFromIndex toValue: to at: physicalToIndex

	| voidStart voidEnd |

	voidStart := (self isStartOfInterval: physicalFromIndex)
		ifTrue:
			[ "Extend to of this gap back a little."
			intervals
				at: physicalFromIndex + 1
				put: from.
			physicalFromIndex + 1 ]
		ifFalse: [ physicalFromIndex ].
	voidEnd := ((self isStartOfInterval: physicalToIndex)
		and: [ physicalToIndex > physicalFromIndex  "account for when to is an endpoint exactly on the end of an interval" ])
			ifTrue:
				[ "Account for removing exact endpoints of an existing interval."
				(intervals at: physicalToIndex + 1) + proximityThreshold = to
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse:
						[ "Extend from of next gap up a little.."
						intervals
							at: physicalToIndex
							put: to.
						physicalToIndex ] ]
			ifFalse: [ physicalToIndex + 1 ].
	self
		voidFrom: voidStart
		to: voidEnd! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
extendIntervalFrom: from at: physicalFromIndex to: to at: physicalToIndex
	
	| bridgeStart bridgeEnd |

	bridgeStart := (self isStartOfGap: physicalFromIndex)
		ifTrue:
			[ "Extend from of next interval back a little.."
			(physicalFromIndex < self physicalSize and: [ (intervals at: physicalFromIndex + 1) <= (to + proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalFromIndex + 1 "start of next interval"
						put: from ].
			physicalFromIndex > 0
				ifTrue: [ (from - proximityThreshold) <= (intervals at: physicalFromIndex)
					ifTrue: [ physicalFromIndex - 1 ]
					ifFalse: [ physicalFromIndex + 1 ] ]
				ifFalse: [ 1 ] ]
		ifFalse:
			[ bridgeStart := physicalFromIndex ].
	bridgeEnd := (self isStartOfGap: physicalToIndex)
		ifTrue:
			[ "Extend to of this interval forward a little.."
			(physicalToIndex > 0 and: [ (intervals at: physicalToIndex) >= (from - proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalToIndex  "end of this interval"
						put: to ].
			physicalToIndex < self physicalSize
				ifTrue: [ (to + proximityThreshold) >= (intervals at: physicalToIndex + 1)
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse: [ physicalToIndex ] ]
				ifFalse: [ self physicalSize ] ]
		ifFalse:
			[ bridgeEnd := physicalToIndex + 1 ].
	self
		bridgeFrom: bridgeStart
		to: bridgeEnd! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
fromValue: aMagnitude isOutsideProximityThresholdAt: physicalFromIndex

	^physicalFromIndex = 0
		or: [ (aMagnitude - proximityThreshold) > (intervals at: physicalFromIndex) ]! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/17/2002 15:43'!
gaps

	| answer |
	answer _ OrderedCollection new.
	self gapsDo:
		[ :eachLow :eachHigh |
		answer add: (eachLow to: eachHigh) ].
	^answer! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
gapsDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock

	self gapsDo:
		[ :eachLow :eachHigh |
		(twoArgBlock
			value: eachLow
			value: eachHigh)
				ifTrue:
					[ ^foundBlock
						value: eachLow
						value: eachHigh ] ].
	^noneBlock value! !

!MaIntervalCollection methodsFor: '-- all --' stamp: 'cmm 8/17/2002 15:47'!
gapsDo: twoArgBlock
	"For each gap in the receivers intervals, value twoArgBlock.
	The first argument is the low of the gap, the second is the high."

	2
		to: intervals size - 2
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex) + proximityThreshold
				value: (intervals at: eachLowIndex + 1) - proximityThreshold ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: 'cmm 8/25/2002 20:45'!
hasGaps

	^intervals size > 2! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
highest

	^intervals isEmpty ifTrue: [ self error: self printString, ' is empty.' ] ifFalse: [ intervals last ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
includes: aMagnitude

	^(self indexOf: aMagnitude) > 0! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 3/24/2005 12:02'!
independentCopy

	^ self copy postIndependentCopy! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
indexOf: aMagnitude
	"Answer the index of the interval which contains aMagnitude.
	If aMagnitude is not part of any interval in the receiver, answer 0."

	| index |

	index := self physicalIndexOf: aMagnitude.
	^((self isStartOfInterval: index)
		and: [ (intervals at: index + 1) > aMagnitude ])
			ifTrue:
				[ self logicalIndexFor: index ]
			ifFalse:
				[ 0 ]! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 7/17/2002 22:53'!
initialize

	super initialize.

	"The intervals is stored in sorted order.  Each odd is the low of an interval, the
	even is the high."
	intervals := OrderedCollection new: 2.

	"proximityThreshold is a value used to determine how close the
	To end of an interval needs to be to the From end of the next
	interval before they will be merged into a single interval.  The
	default is 0 (no merging)."
	proximityThreshold := 0! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/17/2002 15:40'!
intervals
	"This needed only for the MaPersistentObjectsApp.  Normally, we don't need this, but to efficiently update the Magma 'memory' on disk, it is best to have direct access to this."

	^intervals
! !

!MaIntervalCollection methodsFor: 'actions' stamp: ''!
intervalsDo: twoArgBlock
	"For each interval in the receivers intervals, value twoArgBlock.
	The first argument is the low of the interval, the second is the high."

	1
		to: intervals size - 1
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex)
				value: (intervals at: eachLowIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
isEmpty

	^intervals isEmpty! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
isStartOfGap: physicalIndex

	^physicalIndex even! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
isStartOfInterval: physicalIndex

	^physicalIndex odd! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
logicalIndexFor: physicalIndex

	^(physicalIndex + 1) // 2! !

!MaIntervalCollection methodsFor: 'accessing' stamp: 'cmm 8/29/2002 22:45'!
lowest

	^intervals first! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
maDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock 
	self intervalsDo: 
			[:eachLow :eachHigh | 
			(twoArgBlock value: eachLow value: eachHigh) 
				ifTrue: [^foundBlock value: eachLow value: eachHigh]].
	^noneBlock value! !

!MaIntervalCollection methodsFor: 'printing' stamp: 'cmm 2/5/2003 16:33'!
maPrintAbbreviatedOn: aStream

	1
		to: intervals size
		do:
			[ :index |
			aStream nextPutAll: (intervals at: index) printString.
			index odd ifTrue: [ aStream nextPutAll: '..' ] ifFalse: [ aStream nextPutAll: '  ' ] ]
! !

!MaIntervalCollection methodsFor: 'testing' stamp: ''!
notEmpty

	^self isEmpty not
! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalIndexFor: logicalIndex

	^(logicalIndex * 2 - 1) max: 0! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalIndexOf: aMagnitude

	| index |

	index := intervals size.
	[ index > 0 and: [ (intervals at: index) >= aMagnitude ] ] whileTrue: [ index := index - 1 ].
	^index! !

!MaIntervalCollection methodsFor: 'private' stamp: ''!
physicalSize

	^intervals size! !

!MaIntervalCollection methodsFor: 'private creation' stamp: 'cmm 7/1/2004 10:49'!
postCopy

	intervals := intervals copy! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 3/24/2005 12:02'!
postIndependentCopy

	intervals _ intervals copy! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
privateInsertFrom: low to: high afterIndex: physicalIndex

	intervals
		add: high
		afterIndex: physicalIndex.
	intervals
		add: low
		afterIndex: physicalIndex! !

!MaIntervalCollection methodsFor: 'private building' stamp: ''!
proximityThreshold: aMagnitude

	proximityThreshold := aMagnitude! !

!MaIntervalCollection methodsFor: 'building' stamp: ''!
remove: anInterval

	^self
		removeFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building' stamp: 'cmm 8/18/2002 00:59'!
removeFrom: from to: to

	| physicalFromIndex physicalToIndex |

	from = to
		ifTrue:
			[ "User does not realize that this class deals with real-number-based
			intervals that are inclusive on the left and exclusive on the right.
			Take care of him by removing a sensible interval."
			^self
				removeFrom: from
				to: to + proximityThreshold ].
	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "If completely inside an interval, we must add a new interval to break it up"
			((self isStartOfInterval: physicalFromIndex)
				and: [ (intervals at: physicalFromIndex + 1) > to ])
					ifTrue:
						[ self
							privateInsertFrom: from - proximityThreshold
							to: to + proximityThreshold
							afterIndex: physicalFromIndex.
						^from ] ].
	self
		extendGapFromValue: from - proximityThreshold
		at: physicalFromIndex
		toValue: to + proximityThreshold
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'accessing' stamp: ''!
size

	^intervals size / 2! !

!MaIntervalCollection methodsFor: 'private testing' stamp: ''!
toValue: aMagnitude isOutsideProximityThresholdAt: physicalToIndex

	^physicalToIndex = self physicalSize
		or: [ (aMagnitude + proximityThreshold) < (intervals at: physicalToIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 7/3/2002 00:13'!
voidFrom: voidStart to: voidEnd

	voidEnd - voidStart - 1
		timesRepeat: [ intervals removeAt: voidStart + 1 ]! !

MaObject subclass: #MaLargeArrayOfNumbers
	instanceVariableNames: 'file byteSize byteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Integer collections'!

!MaLargeArrayOfNumbers commentStamp: 'cmm 4/28/2005 20:55' prior: 0!
MaLargeArrayOfNumbers is an auto-growing, never-shrinking array of numbers.  All numbers have the same number of bits available to represent them.  It maintains the collection in a single randomly-accessed file, and is therefore constrained by available disk space or the addressability of the file.!

!MaLargeArrayOfNumbers class methodsFor: 'utilities' stamp: 'cmm 2/15/2005 22:44'!
create: filenameString bitSize: anInteger

	^self new
		byteSize: anInteger / 8;
		create: filenameString
! !

!MaLargeArrayOfNumbers class methodsFor: 'creation' stamp: ''!
open: aString

	^self new open: aString
! !

!MaLargeArrayOfNumbers class methodsFor: 'accessing' stamp: 'cmm 3/20/2005 19:23'!
suggestedFileExtension
	^ 'idx'! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: ''!
add: anInteger

	^self
		at: self size + 1
		put: anInteger
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 3/15/2005 21:42'!
at: index

	^ self
		at: index
		ifAbsent: [ self error: 'index out of bounds' ]! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 10/23/2002 23:08'!
at: logicalIndex ifAbsent: aBlock

	(logicalIndex
		between: 1
		and: self size)
			ifFalse: [ ^aBlock value ].
	file
		maRead: byteSize
		bytesFromPosition: 1
		of: byteArray
		atFilePosition: (self filePositionFor: logicalIndex).
	^self valueInByteArray
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 3/15/2005 16:44'!
at: logicalIndex put: anInteger

	logicalIndex > self maxSize ifTrue: [ MaUserError signal: 'Out of range.' ].
	self loadByteArrayWith: anInteger.
	file
		maWrite: byteSize
		bytesFromPosition: 1
		of: byteArray
		atFilePosition: (self filePositionFor: logicalIndex).
	^anInteger
! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: ''!
byteSize: anInteger

	byteSize _ anInteger.
	self initializePremadeByteArray
! !

!MaLargeArrayOfNumbers methodsFor: 'actions' stamp: 'cmm 3/24/2003 23:40'!
close

	file ifNotNil: [ file closed ifFalse: [ file close ] ]
! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 2/15/2005 22:37'!
create: fileNameString

	file _
		(FileStream concreteStream new
			open: fileNameString
			forWrite: true) binary.
	self
		writeHeader ;
		close! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: ''!
filePositionFor: logicalIndex

	^((logicalIndex - 1) * byteSize) + self headerSize
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 3/17/2005 22:41'!
filename

	^ file name! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: ''!
headerSize
	"Header consists of:
		number of bits per Integer in the receiver."
	^1
! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: 'cmm 1/27/2003 23:51'!
initializeFromHeader: aByteArray

	self byteSize: (
		aByteArray 
			maUint: 8
			at: 0)
! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: ''!
initializePremadeByteArray

	byteArray _ ByteArray new: byteSize
! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 1/27/2003 23:58'!
loadByteArrayWith: anInteger

	^byteArray
		maUint: byteSize * 8
		at: 0
		put: anInteger
! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: ''!
loadHeader

	| headerBytes |

	headerBytes _ ByteArray new: self headerSize.
	file
		maRead: self headerSize
		bytesFromPosition: 1
		of: headerBytes
		atFilePosition: 0.
	self initializeFromHeader: headerBytes
! !

!MaLargeArrayOfNumbers methodsFor: 'printing' stamp: 'cmm 2/5/2003 00:25'!
maPrintAbbreviatedOn: aStream

	1
		to: (self size min: 100)
		do:
			[ :each |
			aStream
				space;
				maPrint: (self at: each ifAbsent: [ 'error at: ', each printString ]) ].
	self size > 100 ifTrue: [ aStream nextPutAll: ' ...' ]! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 3/15/2005 21:40'!
maxFileSize

	^ 2000000000! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 3/15/2005 21:40'!
maxSize

	^ self maxFileSize // byteSize! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: 'cmm 1/27/2003 23:58'!
newHeaderBytes
	"Answer a new ByteArray containing the header for the receiver.
	This is only used when creating the an instance of this class with
	the create: constructor."

	^(ByteArray new: self headerSize)
		maUint: 8
		at: 0
		put: byteSize;
		yourself
! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 2/15/2005 15:23'!
open: fileNameString

	file _ (FileStream concreteStream new
		open: fileNameString
		forWrite: true).
	file ifNil: [ MagmaEnvironmentError signal: fileNameString, ' could not be opened.' ].
	file binary.
	self loadHeader
! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: ''!
size

	^(file size - self headerSize) / byteSize! !

!MaLargeArrayOfNumbers methodsFor: 'accessing' stamp: 'cmm 4/13/2005 15:27'!
upTo: countInteger startingAt: startInteger

	| answer size targetX |
	size _ self size.
	startInteger > size ifTrue: [ ^ #() ].
	answer _ Array new: (countInteger min: size-startInteger+1).
	targetX _ 0.
	startInteger to: (startInteger+countInteger-1 min: size) do:
		[ : sourceX |
		answer
			at: (targetX _ targetX + 1)
			put: (self at: sourceX) ].
	^ answer! !

!MaLargeArrayOfNumbers methodsFor: 'private' stamp: 'cmm 1/27/2003 23:51'!
valueInByteArray

	^byteArray
		maUint: byteSize * 8
		at: 0
! !

!MaLargeArrayOfNumbers methodsFor: 'initialize' stamp: ''!
writeHeader

	| headerBytes |

	headerBytes _ self newHeaderBytes.
	file
		maWrite: self headerSize
		bytesFromPosition: 1
		of: headerBytes
		atFilePosition: 0
! !

MaObject subclass: #MaMagnitudeIndex
	instanceVariableNames: 'file keySize valueSize nextRecordNumber highestPossibleKey'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Integer collections'!

!MaMagnitudeIndex commentStamp: '<historical>' prior: 0!
keySize in bits.!

MaMagnitudeIndex subclass: #MaHashIndex
	instanceVariableNames: 'hashRecordSize record recordFactory preMadeKeysArrays preMadeValuesArrays'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Integer collections'!

!MaHashIndex class methodsFor: 'creation' stamp: ''!
create: fileNameString keySize: keySizeInteger

	self shouldNotImplement
! !

!MaHashIndex class methodsFor: 'creation' stamp: 'cmm 4/28/2005 21:15'!
create: fileNameString keySize: numberOfBitsInteger valueSize: valueSize recordSize: entriesInteger

	^self new
		keySize: numberOfBitsInteger ;
		valueSize: valueSize ;
		hashRecordSize: entriesInteger;
		initializeRecords;
		create: fileNameString
! !

!MaHashIndex class methodsFor: 'creation' stamp: ''!
suggestedFileExtension

	^'hdx'
! !

!MaHashIndex methodsFor: 'add / remove' stamp: 'cmm 4/28/2005 21:55'!
add: valueInteger at: keyInteger
	"Find the record which either has space for keyInteger or is as far zoomed-in as possible (with either space for keyInteger or not).  If a record already contains another key greater than keyInteger, that entry must be 'pushed' down so we can maintain key order.  Additionally, we must incremement the numberOfChildren for each record to the path of where we finally add."

	| currentKeyToWrite currentValueToWrite hasSpace reachedBottom |
	self 
		validateCanAdd: valueInteger
		at: keyInteger.
	currentKeyToWrite _ keyInteger.
	currentValueToWrite _ valueInteger.
	self selectRoot.
	hasSpace _ false.
	reachedBottom _ false.
	[ hasSpace | reachedBottom ] whileFalse:
		[ | highestBax keyToMove valueToMove childRecordNumber |
		highestBax _ record highestByteArrayIndexFor: currentKeyToWrite.
		((hasSpace _ record canAdd: currentKeyToWrite)
			or: [ (keyToMove _ record keyAtHighestSlotFor: currentKeyToWrite) > currentKeyToWrite ])
				ifTrue:
					[ hasSpace
						ifFalse:
							[ valueToMove _ record valueAtByteArrayIndex: highestBax.
							record removeEntryAtByteArrayIndex: highestBax ].
					"Incrementing here is 'corrected', below, if necessary.."
					record incrementNumberOfChildrenAtByteArrayIndex:
						(record 
							atKey: currentKeyToWrite
							put: currentValueToWrite).
					hasSpace
						ifFalse:
							[ "It should be safe to assume, here, that highestBax is where currentKeyToWrite was just placed."
							(record canAdd: keyToMove)
								ifTrue:
									[ record
										incrementNumberOfChildrenAtByteArrayIndex:
											(record
												atKey: keyToMove
												put: valueToMove);
										decrementNumberOfChildrenAtByteArrayIndex: highestBax.
									hasSpace _ true ]
								ifFalse:
									[ | newBax |
									"If the displaced entry will not be a child of the displacer, we need to correct the increment, (see) above, by decrement the numberOfChildren and incrementing where it *will* be placed."
									newBax _ record highestByteArrayIndexFor: keyToMove.
									highestBax = newBax
										ifFalse:
											[ record
												decrementNumberOfChildrenAtByteArrayIndex: highestBax ;
												incrementNumberOfChildrenAtByteArrayIndex: newBax ] ] ] ]
				ifFalse:
					[ record incrementNumberOfChildrenAtByteArrayIndex: highestBax ].
		self writeRecord: record.
		valueToMove notNil
			ifTrue:
				[
				"we will need to push this key/value pair down"
				currentKeyToWrite _ keyToMove.
				currentValueToWrite _ valueToMove ].
		childRecordNumber _ record recordNumberCloserTo: currentKeyToWrite.
		reachedBottom _ childRecordNumber = 0.
		reachedBottom ifFalse: [ self selectRecord: childRecordNumber ] ].
	(reachedBottom and: [ hasSpace not ])
		ifTrue:
			[ self selectNewRecordFor: currentKeyToWrite.
			record incrementNumberOfChildrenAtByteArrayIndex:
				(record
					atKey: currentKeyToWrite
					put: currentValueToWrite).
			self writeRecord ]
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 12/5/2002 00:04'!
allRecordsDo: twoArgBlock

	self selectRoot.
	self
		recordsFrom: record independentCopy
		do: twoArgBlock! !

!MaHashIndex methodsFor: 'private' stamp: ''!
create: fileNameString

	super create: fileNameString.
	self writeNewRoot
! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 4/28/2005 21:23'!
createRecord

	record _
		(MaHashIndexRecord
			keySize: keySize
			valueSize: valueSize
			numberOfEntries: hashRecordSize)

				beFrom: 0
				to: (2 raisedTo: keySize) - 1;

				recordNumber: 0
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 3/22/2005 18:05'!
do: aBlock for: aMagmaFileTraverser using: aMaReadStrategy

	self
		entriesFromIndex: 1
		do:
			[ :eachRecord :recordPosition :levels |
			aMagmaFileTraverser
				graphFrom: (eachRecord keyAtByteArrayIndex: recordPosition)
				do: aBlock
				using: aMaReadStrategy ]
		until:
			[ :eachRecord :recordPosition :levels | false ]! !

!MaHashIndex methodsFor: 'debugging' stamp: 'cmm 10/30/2002 21:03'!
dumpEntries

	self dumpEntriesTo: Transcript! !

!MaHashIndex methodsFor: 'debugging' stamp: 'cmm 11/10/2002 20:31'!
dumpEntriesTo: aStream

	^self
		dumpEntriesTo: aStream
		fromIndex: 0
		count: self size! !

!MaHashIndex methodsFor: 'debugging' stamp: 'cmm 4/28/2005 21:54'!
dumpEntriesTo: aStream fromIndex: fromIndex count: anInteger

	| entryNumber originalRecord stopwatch |
	originalRecord _ record recordNumber.
	entryNumber _ fromIndex.
	stopwatch _ MaStopwatch new start.
	MagmaNotification signal: 'about to dump ', self size printString, ' entries.'.
	aStream endEntry.
	aStream nextPutAll: self filename.
	self
		entriesFrom: fromIndex
		do:
			[ : eachRecord : index : level |
			aStream
				cr;
				nextPutAll: 'at: ' , entryNumber printString;
				tab: level - 1;
				nextPutAll: eachRecord recordNumber printString;
				nextPutAll: ' (' , eachRecord lowKey printString;
				nextPutAll: ' to: ' , eachRecord highKey printString;
				nextPutAll: ')';
				nextPutAll: ' x:' , (eachRecord entryIndexForBax: index) printString;
				nextPutAll: ' k:' , (eachRecord keyAtByteArrayIndex: index) printString;
				nextPutAll: ' o:' , (eachRecord valueAtByteArrayIndex: index) printString;
				nextPutAll: ' r:' , (eachRecord recordNumberAtByteArrayIndex: index) printString;
				nextPutAll: ' c:' , (eachRecord numberOfChildrenAtByteArrayIndex: index) printString.
			entryNumber _ entryNumber + 1.
			stopwatch time > 5 maSeconds ifTrue: [ MagmaNotification signal: entryNumber printString, ' records dumped.'.
			stopwatch reset; start. ] ]
		until: [ : eachRecord : index : level | entryNumber - fromIndex > anInteger ].
	self selectRecord: originalRecord.
	aStream endEntry.
	^ aStream! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
entriesFrom: keyInteger do: threeArgBlock until: conditionBlock
	"The first two arguments of threeArgBlock and conditionBlock are the record and the
	internal index position needed to access the of each entry in those records via
	keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.
	The third argument is the levels from the root record, where the root is level 1.
	This is used to reference into the receivers preMadeRecords, a performance optimization."

	^self
		slotsFrom: keyInteger
		do:
			[ :currentRecord :byteArrayIndex :level |
			(currentRecord hasEntryAtByteArrayIndex: byteArrayIndex)
				ifTrue:
					[ threeArgBlock
						value: currentRecord
						value: byteArrayIndex
						value: level ] ]
		until: conditionBlock
! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
entriesFromIndex: globalIndex do: threeArgBlock until: conditionBlock

	^self
		slotsFromIndex: globalIndex
		do:
			[ :currentRecord :byteArrayIndex :level |
			(currentRecord hasEntryAtByteArrayIndex: byteArrayIndex)
				ifTrue:
					[ threeArgBlock
						value: currentRecord
						value: byteArrayIndex
						value: level ] ]
		until: conditionBlock
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
filePositionFor: recordNumberInteger

	^self headerSize + (record bufferSize * recordNumberInteger) + 1
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 3/21/2003 00:51'!
findAnomalies

	| answer |
	answer _ Dictionary new.
	self selectRoot.
	self allRecordsDo:
		[ :eachParentRec :eachChildRec |
		(eachChildRec 
			addAnomaliesTo: answer 
			usingParent: eachParentRec) ifTrue:
				[ MagmaNotification new
					tag: answer;
					signal ] ].
	^answer! !

!MaHashIndex methodsFor: 'private' stamp: ''!
getNextRecordNumber

	^nextRecordNumber _ nextRecordNumber + 1
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 3/21/2003 00:51'!
hasAnomalies

	^[ self findAnomalies.
	false ]
		on: MagmaNotification
		do: [ :exc | exc return: true ]! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
hashRecordSize

	^hashRecordSize! !

!MaHashIndex methodsFor: 'private' stamp: ''!
hashRecordSize: numberOfEntriesInteger

	hashRecordSize _ numberOfEntriesInteger
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
headerSize
	"Header consists of:
		keySize:  16-bits
		recordSize: 16-bits"

	^4
! !

!MaHashIndex methodsFor: 'testing' stamp: ''!
includesKey: anInteger

	self selectFirstRecordWithKeyEqualOrGreaterThan: anInteger.
	^record hasEntryForKey: anInteger
! !

!MaHashIndex methodsFor: 'testing' stamp: 'cmm 4/28/2005 21:55'!
includesValue: valueInteger at: keyInteger

	| hasEntry key |
	self
		entriesFrom: keyInteger
		do:
			[ : eachRecord : eachByteArrayIndex : level |
			(hasEntry and:
				[ key = keyInteger
					and: [ (eachRecord valueAtByteArrayIndex: eachByteArrayIndex) = valueInteger ] ])
						ifTrue:
							[ ^ true ] ]
		until:
			[ : eachRecord : eachByteArrayIndex : level |
			"Until we're on an entry that could not possibly hold keyInteger."
			((hasEntry _ eachRecord hasEntryAtByteArrayIndex: eachByteArrayIndex)
				and: [ (key _ eachRecord keyAtByteArrayIndex: eachByteArrayIndex) > keyInteger ])
					or: [ eachByteArrayIndex > (eachRecord highestByteArrayIndexFor: keyInteger) ] ].
	^ false! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
indexOf: keyInteger

	^self
		indexOfLowest: keyInteger
		exactKey: true
! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 12/3/2002 23:13'!
indexOfHighest: keyInteger
	"If keyInteger is present, answer the index of its last occurrence.
	Otherwise, answer the index of the next lower key.  If keyInteger
	is lower than all keys in the receiver, answer 0."

	| lowestIndexOfNextHigherKey |

	keyInteger >= self highestPossibleKey ifTrue: [ ^self size ].
	lowestIndexOfNextHigherKey _
		self
			indexOfLowest: keyInteger + 1
			exactKey: false.
	lowestIndexOfNextHigherKey = 1 ifTrue: [ ^0 ]. "lower than all"
	^lowestIndexOfNextHigherKey = 0
				ifTrue:
					[ self size ]
				ifFalse:
					[ lowestIndexOfNextHigherKey - 1 max: 1 ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/30/2002 13:20'!
indexOfLowest: keyInteger exactKey: aBoolean
	"If aBoolean is true, answer 0 if we don't find an exact key.  If false
	answer the index of the first equal or higher key."

	| total index |

	total _ 0.
	self
		navigateToward: keyInteger
		until:
			[ :currentRecord | | foundRecord |
			foundRecord _ false.
			currentRecord
				eligibleByteArrayIndexesFor: keyInteger
				do:
					[ :each |
					((currentRecord hasEntryAtByteArrayIndex: each)
						and: [ (currentRecord keyAtByteArrayIndex: each) >= keyInteger ])
							ifTrue:
								[ foundRecord _ true ] ].
			total _ total + (record numberOfChildrenBefore: keyInteger).
			foundRecord ]
		do: [ :currentRecord | "nada" ].
	"We now have the correct potential record selected and a correct total number
	of entries that occur before the entry that would contain keyInteger.  Now determine
	whether we should answer it or nil"
	index _ total + 1.
	^aBoolean
		ifTrue:
			[ (record hasEntryForKey: keyInteger) ifTrue: [ ^index ] ifFalse: [ 0 ] ]
		ifFalse:
			[ index > self size ifTrue: [ 0 ] ifFalse: [ index ] ]! !

!MaHashIndex methodsFor: 'initializing' stamp: ''!
initialize

	super initialize.
	self initializePreMadeAnswerArrays
! !

!MaHashIndex methodsFor: 'initializing' stamp: 'cmm 4/17/2005 22:01'!
initializePreMadeAnswerArrays

	preMadeKeysArrays _ (1 to: 250) collect: [ :eachSize | Array new: eachSize ].
	preMadeValuesArrays _ (1 to: 250) collect: [ :eachSize | Array new: eachSize ]! !

!MaHashIndex methodsFor: 'initializing' stamp: ''!
initializeRecordFactory

	recordFactory _ MaHashIndexRecordFactory index: self
! !

!MaHashIndex methodsFor: 'initializing' stamp: ''!
initializeRecords

	self
		createRecord;
		initializeRecordFactory
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/17/2005 22:02'!
keysArrayOfSize: anInteger

	^anInteger > preMadeKeysArrays size
		ifTrue:
			[ Array new: anInteger ]
		ifFalse:
			[ preMadeKeysArrays at: anInteger ]
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 1/27/2003 23:50'!
loadHeaderFrom: aByteArray

	self
		keySize: (aByteArray maUint: 16 at: 0);
		hashRecordSize: (aByteArray maUint: 16 at: 2);
		initializeRecords
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/27/2002 13:22'!
navigateToward: keyInteger until: untilBlock do: doBlock
	"Select records between the root and the one that can contain keyInteger
	and value one-argument doBlock on each of those records until the
	one-argument untilBlock values true or we cannot navigate any
	closer to keyInteger."

	| level |

	level _ 1.
	self selectRoot.
	[ | goalRecordReached bax |
	bax _ record highestByteArrayIndexFor: keyInteger.
	recordFactory
		put: record
		atLevel: level
		byteArrayIndex: bax.
	goalRecordReached _ untilBlock value: record.
	goalRecordReached not and: 	[ record hasChildAtByteArrayIndex: bax ] ]
		whileTrue:
			[ doBlock value: record.
			self selectRecord: (record recordNumberCloserTo: keyInteger).
			level _ level + 1 ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 12/4/2002 23:30'!
numberOfEntries

	^record numberOfEntries! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
numberOfEntriesFrom: lowKey to: highKey
	"Inclusive of lowKey and highKey, if those keys exist.
	If lowKey does not exist, we count from the next higher key.
	If highKey doesn't exist, we count up to the next highestKey."

	| lowest |

	(lowest _
		self
			indexOfLowest: lowKey
			exactKey: false) = 0
				ifTrue:
					[ ^0 ].
	^(self indexOfHighest: highKey) - lowest + 1
! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
occurrencesOf: keyInteger

	^self
		numberOfEntriesFrom: keyInteger
		to: keyInteger
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/28/2005 21:18'!
populateHeader: aByteArray

	aByteArray
		maUint: 8
		at: 0
		put: keySize.
	aByteArray
		maUint: 8
		at: 1
		put: valueSize.
	aByteArray
		maUint: 8
		at: 3
		put: hashRecordSize! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 12/25/2002 12:51'!
readHeader

	| header |

	header _ ByteArray new: self headerSize.
	file
		maRead: self headerSize
		bytesFromPosition: 1
		of: header
		atFilePosition: 0.
	self
		loadHeaderFrom: header;
		initializeRecords! !

!MaHashIndex methodsFor: 'private' stamp: ''!
readRecord: aMaHashIndexRecord at: recordIndex

	aMaHashIndexRecord
		readFrom: file
		atPosition: (self filePositionFor: recordIndex);

		recordNumber: recordIndex.
	^aMaHashIndexRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
recordSize

	^record bufferSize
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 11/30/2002 00:36'!
recordsDo: twoArgBlock level: level byteArrayIndex: anInteger

	| parent |
	recordFactory
		put: record
		atLevel: level
		byteArrayIndex: anInteger.
	parent _ recordFactory recordAtLevel: level.
	twoArgBlock value: parent value: level.
	parent
		byteArrayIndexesFrom: 0
		do:
			[ :eachBax |
			(parent hasChildAtByteArrayIndex: eachBax)
				ifTrue:
					[ self selectRecord: (parent recordNumberAtByteArrayIndex: eachBax).
					self
						recordsDo: twoArgBlock
						level: level + 1
						byteArrayIndex: anInteger ] ]! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 12/5/2002 00:04'!
recordsFrom: aMaHashIndexRecord do: twoArgBlock

	| childRec |
	aMaHashIndexRecord
		byteArrayIndexesFrom: 0
		do:
			[ :eachBax |
			(aMaHashIndexRecord hasChildAtByteArrayIndex: eachBax)
				ifTrue:
					[ childRec _ aMaHashIndexRecord independentCopy.
					self 
						readRecord: childRec
						at: (aMaHashIndexRecord recordNumberAtByteArrayIndex: eachBax).
					twoArgBlock
						value: aMaHashIndexRecord
						value: childRec.
					self
						recordsFrom: childRec
						do: twoArgBlock ] ]! !

!MaHashIndex methodsFor: 'add / remove' stamp: 'cmm 4/28/2005 21:54'!
remove: valueInteger at: keyInteger
	"Answer a boolean indicating whether anything was removed."

	| currentBaxToFill sourceBax parentRecord childRecord baxFilled |
	(self
		includesValue: valueInteger
		at: keyInteger)
			ifFalse:
				[ ^false ].
	self
		navigateToward: keyInteger
		until:
			[ :eachRecord |
			(currentBaxToFill _ eachRecord
				removeKey: keyInteger
				withValue: valueInteger)
					notNil ]
		do:
			[ :eachRecord |
			eachRecord decrementNumberOfChildrenAtKey: keyInteger.
			self writeRecord ].
	self writeRecord.

	"get out if we can.."
	currentBaxToFill ifNil: [ ^ true ].
	(record hasChildAtByteArrayIndex: currentBaxToFill) ifFalse: [ ^ true ].

	"Now pull up child record entries into the newly vacated slot because add looks for the first available slot that is eligible for the key it is adding, which *could* be greater than the key of the next child (meaning we'd then be out of order)."
	sourceBax _ 0. "non-nil"
	parentRecord _ record independentCopy.
	childRecord _ record independentCopy.
	[ sourceBax notNil and: [ parentRecord hasChildAtHighestByteArrayIndexForBax: currentBaxToFill ] ]
		whileTrue:
			[ | highestByteArrayIndexForBaxToFill |
			highestByteArrayIndexForBaxToFill _ parentRecord highestByteArrayIndexForAnyKeyAt: currentBaxToFill.
			self
				readRecord: childRecord
				at: (parentRecord recordNumberAtByteArrayIndex: highestByteArrayIndexForBaxToFill).
			sourceBax _ childRecord byteArrayIndexOfLowestEntry.
			sourceBax ifNotNil:
				[ baxFilled _ parentRecord
"what if we changed atKey:put: to add to the *highest* rather than lowest possible slot."
					atKey: (childRecord keyAtByteArrayIndex: sourceBax)
					put: (childRecord valueAtByteArrayIndex: sourceBax).
				baxFilled ~= highestByteArrayIndexForBaxToFill
					ifTrue:
						[ parentRecord incrementNumberOfChildrenAtByteArrayIndex: baxFilled.
						parentRecord decrementNumberOfChildrenAtByteArrayIndex: highestByteArrayIndexForBaxToFill ].
				childRecord
					decrementNumberOfChildrenAtByteArrayIndex: sourceBax;
					removeEntryAtByteArrayIndex: sourceBax.
				self
					writeRecord: parentRecord;
					writeRecord: childRecord.
				parentRecord populateFrom: childRecord.
				currentBaxToFill _ sourceBax ] ].
	^true! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectFirstRecordWithKeyEqualOrGreaterThan: keyInteger
	"If keyInteger is
	larger than every key in the index, answer nil, otherwise answer the level of the
	record with the answer."

	| level searchedLevel |

	level _ 1.
	self
		navigateToward: keyInteger
		until:
			[ :currentRecord |
			currentRecord
				eligibleByteArrayIndexesFor: keyInteger
				do:
					[ :each |
					((currentRecord hasEntryAtByteArrayIndex: each)
						and: [ (currentRecord keyAtByteArrayIndex: each) >= keyInteger ])
							ifTrue:
								[ ^level ] ].
			false ]
		do: [ :currentRecord | level _ level + 1 ].
	"Though we have arrived at the record which would contain keyInteger,
	in case it does not have entries whose keys are >= keyInteger,
	look for the next higher key."
	self
		traverseSlotsFromRecord: record
		index: (record entryIndexFor: keyInteger)
		do: [ :currentRecord :byteArrayIndex :currentLevel | "nothing" ]
		until:
			[ :currentRecord :byteArrayIndex :currentLevel |
			searchedLevel _ currentLevel.
			currentRecord hasEntryAtByteArrayIndex: byteArrayIndex ]
		level: level.
	^searchedLevel
! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 12/6/2002 12:22'!
selectIndependentRecord: recordNumber

	| r |
	r _ record independentCopy.
	self readRecord: r at: recordNumber.
	^r! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectNewRecordFor: keyInteger
	"We need to make sure we reference the new record from the
	current record."

	self setUpNewRecordFor: keyInteger.
	"Reference new record and write before selecting the new record."
	record
		recordNumberAtSlotForKey: keyInteger
		put: recordFactory preMadeNewRecord recordNumber.
	self
		writeRecord; "replace record that now references newRecord.."
		selectPreMadeNewRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectPreMadeNewRecord

	record populateFrom: recordFactory preMadeNewRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectRecord: anInteger

	self
		readRecord: record
		at: anInteger
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/30/2002 13:03'!
selectRecordAt: globalIndex
	"Answer the level and the byteArrayIndex of the selected record, or nil if out of bounds.
	Since the caller of this method is going to call traverseSlotsFromRecord:index:do:until:level:,
	we must be sure to populate the receivers path."

	| level byteArrayIndex currentGlobalPosition outOfBounds |

	self selectRoot.
	level _ 1.
	currentGlobalPosition _ 0.  "0-based"
	outOfBounds _ false.
	[
		byteArrayIndex _
			record
				byteArrayIndexContainingGlobalIndex: globalIndex
				base: currentGlobalPosition.
		byteArrayIndex isNil
			ifTrue:
				[ outOfBounds _ true.
				false ]
			ifFalse:
				[ currentGlobalPosition _
					currentGlobalPosition +
						(record numberOfChildrenBeforeByteArrayIndex: byteArrayIndex).
				(record hasEntryAtByteArrayIndex: byteArrayIndex)
					ifTrue: [ currentGlobalPosition _ currentGlobalPosition + 1 ].
				(currentGlobalPosition < globalIndex) and:
					[ "this should never be false if above is true"
					record hasChildAtByteArrayIndex: byteArrayIndex ] ] ]
						whileTrue:
							[ recordFactory
								put: record
								atLevel: level
								byteArrayIndex: byteArrayIndex.
							self selectRecord:
								(record recordNumberAtByteArrayIndex: byteArrayIndex).
							level _ level + 1 ].
	^outOfBounds
		ifTrue:
			[ nil ]
		ifFalse:
			[ Array
				with: byteArrayIndex
				with: level ]! !

!MaHashIndex methodsFor: 'private' stamp: ''!
selectRoot

	self selectRecord: 0
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
setUpNewRecordFor: keyInteger

	recordFactory preMadeNewRecord
		beZoomedInToward: keyInteger
			from: record;
		recordNumber: self getNextRecordNumber
! !

!MaHashIndex methodsFor: 'accessing' stamp: ''!
size

	self selectRoot.
	^record numberOfChildren
! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
slotsFrom: keyInteger do: threeArgBlock until: conditionBlock
	"The first two arguments of threeArgBlock and conditionBlock are the record and the
	internal index position needed to access the of each entry in those records via
	keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.
	The third argument is the levels from the root record, where the root is level 1.
	This is used to reference into the receivers preMadeRecords, a performance optimization."

	| level startingEntryIndex |

	(level _ self selectFirstRecordWithKeyEqualOrGreaterThan: keyInteger) isNil ifTrue: [ ^self ].
	"For inverted records, it is possible the entryIndex for keyInteger will be consumed by the
	next lower key.  Unfortunately, we have to check."
	startingEntryIndex _ record entryIndexFor: keyInteger.
	(record keyAtEntryIndex: startingEntryIndex) < keyInteger
		ifTrue:
			[ startingEntryIndex _ startingEntryIndex + 1 ].
	self
		traverseSlotsFromRecord: record
		index: startingEntryIndex
		do: threeArgBlock
		until: conditionBlock
		level: level
! !

!MaHashIndex methodsFor: 'enumerating' stamp: ''!
slotsFromIndex: globalIndex do: threeArgBlock until: conditionBlock
	"The first two arguments of threeArgBlock and conditionBlock are the record and the
	internal index position needed to access the of each entry in those records via
	keyAtByteArrayIndex:, oidAtByteArrayIndex:, and recordNumberAtByteArrayIndex:.
	The third argument is the levels from the root record, where the root is level 1.
	This is used to reference into the receivers preMadeRecords, a performance optimization."

	| byteArrayIndexAndLevelArray |

	(byteArrayIndexAndLevelArray _ self selectRecordAt: globalIndex) isNil ifTrue: [ ^self ].
	self
		traverseSlotsFromRecord: record
		index: (record entryIndexForBax: byteArrayIndexAndLevelArray first)
		do: threeArgBlock
		until: conditionBlock
		level: byteArrayIndexAndLevelArray last
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/29/2002 22:46'!
traverseSlotsFromParentRecord: aMaHashIndexRecord index: entryIndex do: threeArgBlock until: conditionBlock level: levelInteger
	"levelInteger is 1 for the root record."

	| conditionMet |

	conditionMet _ false.
	aMaHashIndexRecord
		byteArrayIndexesFrom: entryIndex
		do:
			[ :index | | nextToRead |
			conditionMet
				ifFalse:
					[ nextToRead _ aMaHashIndexRecord recordNumberAtByteArrayIndex: index.
					(conditionBlock
						value: aMaHashIndexRecord
						value: index
						value: levelInteger)
							ifTrue:
								[ conditionMet _ true ]
							ifFalse:
								[ threeArgBlock
									value: aMaHashIndexRecord
									value: index
									value: levelInteger ] ].
			conditionMet
				ifFalse:
					[ nextToRead > 0
						ifTrue:
							[ | currentRecord |
							currentRecord _ recordFactory recordAtLevel: levelInteger + 1.
							self
								readRecord: currentRecord
								at: nextToRead.
							conditionMet _
								self
									traverseSlotsFromParentRecord: currentRecord
									index: 0
									do: threeArgBlock
									until: conditionBlock
									level: levelInteger + 1 ] ] ].
	^conditionMet
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/29/2002 22:46'!
traverseSlotsFromRecord: aMaHashIndexRecord index: entryIndex do: threeArgBlock until: conditionBlock level: levelInteger
	"levelInteger is 1 for the root record."

	| done |

	done _
		self
			traverseSlotsFromParentRecord: aMaHashIndexRecord
			index: entryIndex
			do: threeArgBlock
			until: conditionBlock
			level: levelInteger.
	(done not and: [ levelInteger > 1 ])
		ifTrue:
			[ | nextStart |
			nextStart _ recordFactory recordAtLevel: levelInteger - 1.
			self
				readRecord: nextStart
				at: (recordFactory recordNumberAtLevel: levelInteger - 1).
			self
				traverseSlotsFromRecord: nextStart
				index: (recordFactory entryIndexAtLevel: levelInteger - 1) + 1
				do: threeArgBlock
				until: conditionBlock
				level: levelInteger - 1 ]
! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 4/17/2005 22:13'!
upTo: countInteger at: keyInteger

	^ self
		upTo: countInteger
		keysAndValuesFrom: keyInteger
		to: keyInteger
		startingAt: 1! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 4/28/2005 21:52'!
upTo: countInteger keysAndValuesFrom: lowKey to: highKey startingAt: index
	"Answer the specified number of values starting at logical-position index *within* the lowKey to highKey range."

	| lowest answerIndex keys values |
	lowest _ self indexOfLowest: lowKey exactKey: false.
	keys _ self keysArrayOfSize: countInteger.
	values _ self valuesArrayOfSize: countInteger.
	answerIndex _ 0.
	self
		entriesFromIndex: lowest + index - 1
		do:
			[ : currentRecord : byteArrayIndex : level |
			keys
				at: (answerIndex _ answerIndex + 1)
				put: (currentRecord keyAtByteArrayIndex: byteArrayIndex).
			values
				at: answerIndex
				put: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ]
		until:
			[  : currentRecord : byteArrayIndex : level |
			answerIndex = countInteger
			or: [ (currentRecord keyAtByteArrayIndex: byteArrayIndex) > highKey ] ].
	^ answerIndex = 0
		ifTrue: [ { #(). #() } ]
		ifFalse:
			[ answerIndex < countInteger
				ifTrue:
					[ { keys copyFrom: 1 to: answerIndex. 
					values copyFrom: 1 to: answerIndex } ]
				ifFalse: [ { keys. values } ] ]! !

!MaHashIndex methodsFor: 'accessing' stamp: 'cmm 4/28/2005 21:52'!
upTo: countInteger keysAndValuesFromIndex: lowIndex

	| answerIndex keys values |
	keys _ self keysArrayOfSize: countInteger.
	values _ self valuesArrayOfSize: countInteger.
	answerIndex _ 0.
	self
		entriesFromIndex: lowIndex
		do:
			[ : currentRecord : byteArrayIndex :level |
			keys
				at: (answerIndex _ answerIndex + 1)
				put: (currentRecord keyAtByteArrayIndex: byteArrayIndex).
			values
				at: answerIndex
				put: (currentRecord valueAtByteArrayIndex: byteArrayIndex) ]
		until: [ : currentRecord : index : level | answerIndex = countInteger ].
	^ answerIndex = 0
		ifTrue: [ { #(). #() } ]
		ifFalse:
			[ answerIndex < countInteger
				ifTrue:
					[ { keys copyFrom: 1 to: answerIndex.
					values copyFrom: 1 to: answerIndex } ]
				ifFalse:
					[ { keys. values } ] ]! !

!MaHashIndex methodsFor: 'SUnit tests' stamp: 'cmm 4/17/2005 22:03'!
upTo: countInteger keysFrom: startKey

	| result answerIndex |

	result _ self valuesArrayOfSize: countInteger.
	answerIndex _ 1.
	self
		entriesFrom: startKey
		do:
			[ :currentRecord :index :level |
			(currentRecord hasEntryAtByteArrayIndex: index)
				ifTrue:
					[ result
						at: answerIndex
						put: (currentRecord keyAtByteArrayIndex: index).
					answerIndex _ answerIndex + 1 ] ]
		until: [ :currentRecord :index :level | answerIndex > countInteger ].
	"Need to check in case of a small index (say only 9 entries?)."
	^answerIndex < countInteger
		ifTrue:
			[ result
				copyFrom: 1
				to: answerIndex - 1 ]
		ifFalse:
			[ result ]
! !

!MaHashIndex methodsFor: 'validation' stamp: 'cmm 3/20/2005 17:49'!
validateCanAdd: oidInteger at: keyInteger

	keyInteger > self highestPossibleKey ifTrue:
		[ MaHashIndexUserError signal: keyInteger printString, ' is larger than the highestPossibleKey.' ]! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 4/17/2005 22:01'!
valuesArrayOfSize: anInteger

	^anInteger > preMadeValuesArrays size
		ifTrue:
			[ Array new: anInteger ]
		ifFalse:
			[ preMadeValuesArrays at: anInteger ]
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
writeHeader

	| header |

	header _ ByteArray new: self headerSize.
	self populateHeader: header.
	file
		maWrite: self headerSize
		bytesFromPosition: 1
		of: header
		atFilePosition: 0
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
writeNewRoot

	self writeRecord
! !

!MaHashIndex methodsFor: 'private' stamp: ''!
writeRecord

	self writeRecord: record
! !

!MaHashIndex methodsFor: 'private' stamp: 'cmm 11/17/2002 22:45'!
writeRecord: aMaHashIndexRecord

	aMaHashIndexRecord
		writeTo: file
		atPosition: (self filePositionFor: aMaHashIndexRecord recordNumber)! !

!MaMagnitudeIndex class methodsFor: 'creation' stamp: ''!
create: aString keySize: numberOfBitsInteger

	^self new
		keySize: numberOfBitsInteger;
		create: aString
! !

!MaMagnitudeIndex class methodsFor: 'creation' stamp: ''!
open: fileNameString

	^self new open: fileNameString
! !

!MaMagnitudeIndex methodsFor: 'building' stamp: ''!
add: oidInteger at: keyInteger

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'actions' stamp: 'cmm 3/6/2003 21:42'!
close

	file ifNotNil: [ file closed ifFalse: [ file close ] ]
! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 3/11/2005 14:39'!
create: fileNameString

	(FileStream concreteStream isAFileNamed: fileNameString)
		ifTrue:
			[ MagmaEnvironmentError signal: fileNameString, ' already exists.' ].
	file _
		(FileStream concreteStream new
			open: fileNameString
			forWrite: true) binary.
	self writeHeader.
	nextRecordNumber _ 0! !

!MaMagnitudeIndex methodsFor: 'actions' stamp: 'cmm 11/26/2002 21:12'!
delete

	file closed ifFalse: [ file close ].
	(FileDirectory forFileName: file name) deleteFileNamed: file name! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 8/22/2002 19:27'!
filename

	^file name! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
headerSize

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 3/20/2005 17:50'!
highestPossibleKey

	^ highestPossibleKey ifNil:
		[ highestPossibleKey _ (2 raisedTo: keySize) - 1 ]
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
keySize

	^keySize! !

!MaMagnitudeIndex methodsFor: 'private' stamp: ''!
keySize: numberOfBitsInteger

	keySize _ numberOfBitsInteger
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
lowestPossibleKey

	^0
! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 10/31/2002 19:37'!
open

	self open: self filename! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 2/15/2005 15:22'!
open: fileNameString

	(FileStream concreteStream isAFileNamed: fileNameString)
		ifFalse:
			[ MagmaEnvironmentError signal: 'Could not locate index file ', fileNameString ].
	file _
		(FileStream concreteStream new
			open: fileNameString
			forWrite: true).
	file ifNil: [ MagmaEnvironmentError signal: fileNameString, ' could not be opened.' ].
	file binary.
	self readHeader.
	nextRecordNumber _ (file size - self headerSize - 1) / self recordSize.
	nextRecordNumber isInteger
		ifFalse:
			[ MagmaSoftwareError signal: 'Invalid index size.  Requires maintenance tool.' ].
	^self! !

!MaMagnitudeIndex methodsFor: 'private' stamp: ''!
readHeader
	"Answer a ByteArray containing the header bytes.  This includes the keySize as well
	as other implementation-specific variables."

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: ''!
recordSize

	self subclassResponsibility
! !

!MaMagnitudeIndex methodsFor: 'framework' stamp: 'cmm 12/23/2002 22:16'!
size

	self subclassResponsibility! !

!MaMagnitudeIndex methodsFor: 'framework' stamp: 'cmm 12/23/2002 22:15'!
upTo: countInteger oidsFrom: lowKey to: highKey

	self subclassResponsibility! !

!MaMagnitudeIndex methodsFor: 'accessing' stamp: 'cmm 4/28/2005 21:12'!
valueSize

	^ valueSize! !

!MaMagnitudeIndex methodsFor: 'private' stamp: 'cmm 4/28/2005 21:16'!
valueSize: numberOfBitsInteger

	valueSize _ numberOfBitsInteger! !

!MaMagnitudeIndex methodsFor: 'private' stamp: ''!
writeHeader
	"Your keySize should be stored somewhere in the header."

	self subclassResponsibility
! !

MaObject subclass: #MaPermutator
	instanceVariableNames: 'arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections'!

!MaPermutator class methodsFor: 'actions' stamp: ''!
permutationsOf: anArrayOfArrays do: aBlock
	"For each combination of elements in anArrayOfArray's, value aBlock
	with the arguments of each combination of elements.  There must be
	as many arguments in aBlock as there are elements in the outer array
	of Array's.

	Note, this also works with any sequenceable collection of sequenceable
	collections."

	| numberOfCombinations args |

	numberOfCombinations :=
		anArrayOfArrays
			inject: 1
			into: [ :prod :each | each size * prod ].
	args := Array new: anArrayOfArrays size.
	1
		to: numberOfCombinations
		do:
			[ :elementIndex |
			1
				to: anArrayOfArrays size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := anArrayOfArrays at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (anArrayOfArrays at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			aBlock valueWithArguments: args ]
! !

!MaPermutator methodsFor: 'building' stamp: ''!
addArguments: aSequenceableCollection

	^self arguments add: aSequenceableCollection
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
arguments

	^arguments! !

!MaPermutator methodsFor: 'building' stamp: ''!
arguments: aSequenceableCollectionOfSequenceableCollections

	arguments := aSequenceableCollectionOfSequenceableCollections
! !

!MaPermutator methodsFor: 'private initializing' stamp: 'cmm 12/24/2002 15:51'!
initialize
	super initialize.
	arguments := OrderedCollection new! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsDo: oneArgBlock
	"For each combination of elements in the receivers args, value aBlock
	with the arguments of each combination of elements.  The argument to
	oneArgBlock is the Array of parameters for each combination."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			oneArgBlock value: args ]
! !

!MaPermutator methodsFor: 'actions' stamp: ''!
permutationsWhere: conditionBlock do: oneArgBlock
	"Same as permutationsDo: except the conditionBlock must also be
	satisfied for aBlock to run.  This is only for mildly better looking code
	since a case statement could achieve the same effect."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			(conditionBlock value: args) ifTrue: [ oneArgBlock value: args ] ]
! !

!MaPermutator methodsFor: 'accessing' stamp: ''!
size

	^self arguments
		inject: 1
		into: [ :prod :each | each size * prod ]
! !

MaObject subclass: #MaTree
	instanceVariableNames: 'parents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma special collections-Tree'!

!MaTree class methodsFor: 'creation' stamp: ''!
new

	^super new initialize
! !

!MaTree methodsFor: 'building' stamp: ''!
add: anObject
	"Adds anObject as a main parent in the receiver."

	self
		add: anObject
		asChildOf: nil.
	^anObject
! !

!MaTree methodsFor: 'building' stamp: 'cmm 2/12/2004 00:02'!
add: anObject asChildOf: aParentObject

	parents at: anObject put: aParentObject.
	^anObject! !

!MaTree methodsFor: 'building' stamp: ''!
addAll: aCollection

	"Adds all in aCollection as a main parents in the receiver."

	aCollection do: [ :each | self add: each ]
! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 22:59'!
addPath: anOrderedCollection

	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  Any parts of the path that need created will be."

	^self
		addPath: anOrderedCollection
		to: nil
! !

!MaTree methodsFor: 'building' stamp: ''!
addPath: anOrderedCollection to: parentObject
	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  If part of the path already exists, there it is
	replaced (though they should be equivalent anyway)."

	anOrderedCollection
		inject: parentObject
		into:
			[ :parent :each |
			self
				add: each
				asChildOf: parent.
			each ].
	^anOrderedCollection
! !

!MaTree methodsFor: 'building' stamp: ''!
addTree: aMaTree

	^self addTree: aMaTree to: nil
! !

!MaTree methodsFor: 'building' stamp: 'cmm 12/24/2002 16:08'!
addTree: aMaTree to: anObject

	(self includes: anObject) ifFalse: [ self error: anObject printString, ' is not part of ', self printString ].
	aMaTree withParentsDo: [ :each :eachParent | self add: each asChildOf: eachParent ].
	^aMaTree
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
allLeafsDo: aBlock
	"Value aBlock with each element in the receiver that has no children."

	| children |

	children := self children.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children includesKey: eachChild) ifFalse: [ aBlock value: eachChild ] ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
allMembers
	"Returns a Set of all the members of the hierarchy"

	^parents keys! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:04'!
allMembersDo: aBlock
	"Returns a Set of all the members of the hierarchy"

	^parents keysDo: aBlock! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 1/23/2002 14:30'!
children

	| children |

	children := Dictionary new: parents size.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children
			at: eachParent
			ifAbsentPut: [ OrderedCollection new ])
				add: eachChild ].
	^children
! !

!MaTree methodsFor: 'accessing' stamp: ''!
childrenOf: anObject 
	^parents maAllKeysAtValue: anObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
collect: aBlock
	"Returns a new tree whose nodes have all been operated on by aBlock."

	^self
		collect: aBlock
		into: self class new
! !

!MaTree methodsFor: 'private' stamp: 'cmm 1/23/2002 14:30'!
collect: aBlock into: aMaTree 
	"Since some objects may share a parent, we use the map to
	make sure we only value each parent just once, otherwise we
	could have some identity problems."

	| map |
	map := Dictionary new.
	parents keysAndValuesDo: 
			[:eachKey :eachValue | 
			eachValue isNil 
				ifTrue: [aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])]
				ifFalse: 
					[aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])
						asChildOf: (map at: eachValue ifAbsentPut: [aBlock value: eachValue])]].
	^aMaTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
conform: aBlock 
	parents keysDo: [:each | (aBlock value: each) ifFalse: [^false]].
	^true! !

!MaTree methodsFor: 'building' stamp: 'cmm 3/24/2005 11:06'!
copy

	^ super copy
		parents: parents ;
		yourself! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock
	"Value twoArgBlock with each object in the receiver and its parent in the order
	of depthFirst."

	self roots do:
		[ :each |
		twoArgBlock
			value: each
			value: nil.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
depthFirstDo: twoArgBlock at: anObject

	(self childrenOf: anObject) do:
		[ :each |
		twoArgBlock
			value: each
			value: anObject.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
depthOf: anObject
	"Returns the number of parents anObject has."

	| depth prevChild |

	(self includes: anObject)
		ifFalse:
			[ self error: self printString , ' does not include ' , anObject printString ].
	depth := 0.
	prevChild := anObject.
	[ prevChild isNil ] whileFalse:
		[ prevChild := self parentOf: prevChild.
		depth := depth + 1 ].
	^depth
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock

	^parents detect: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:02'!
detect: aBlock ifNone: exceptionBlock

	^parents
		detect: aBlock
		ifNone: exceptionBlock value
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:07'!
do: oneArgBlock

	parents keysDo: oneArgBlock! !

!MaTree methodsFor: 'testing' stamp: 'cmm 7/25/2002 19:43'!
hasChildFor: anObject 
	"Returns true if anObject has any children.  Signal an error
	if anObject is not part of the hierarchy."

	^parents includes: anObject! !

!MaTree methodsFor: 'testing' stamp: ''!
hasParentFor: anObject
	"Returns true if anObject has a parent"

	^(self parentOf: anObject) notNil
! !

!MaTree methodsFor: 'testing' stamp: ''!
includes: anObject 
	^parents includesKey: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
initialize
	super initialize.
	self parents: Dictionary new! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 6/10/2003 23:02'!
maDetect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	^parents
		maDetect: aBlock
		ifFound: foundBlock
		ifNone: exceptionBlock value! !

!MaTree methodsFor: 'building' stamp: 'cmm 6/10/2003 23:00'!
move: anObject to: newParentObject 
	^parents at: anObject put: newParentObject! !

!MaTree methodsFor: 'accessing' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject
	"Returns a collection of nodes between someChildObject
	and someParentObject."

	|answer node|

	answer := OrderedCollection new.
	self
		nodesInPathFrom: someChildObject
		to: someParentObject
		do: [ :each | answer add: each ].
	^answer
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
nodesInPathFrom: someChildObject to: someParentObject do: oneArgBlock
	"For each node in the tree from someChildObject to someParentObject, value
	oneArgBlock.  The highest level parent (nil) will not be used as an argument
	to value oneArgBlock."

	| node |

	node := someChildObject.
	oneArgBlock value: someChildObject.
	[ node isNil or: [ node = someParentObject ] ] whileFalse:
		[ oneArgBlock value: node.
		node := self parentOf: node ].
	someParentObject notNil ifTrue: [ oneArgBlock value: someParentObject ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
parentOf: anObject 
	^parents at: anObject! !

!MaTree methodsFor: 'private' stamp: 'cmm 3/24/2005 11:07'!
parents: aDictionary

	parents _ aDictionary! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathFrom: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject ordered from an Object to the top of the hierarchy."

	^self
		nodesInPathFrom: anObject
		to: nil
! !

!MaTree methodsFor: 'enumerating' stamp: ''!
pathFrom: anObject do: aBlock
	"Value aBlock with the objects from anObject to its highest level parent."

	^self
		nodesInPathFrom: anObject
		to: nil
		do: aBlock
! !

!MaTree methodsFor: 'accessing' stamp: ''!
pathTo: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject, ordered from the the top
	of the hierarchy to anObject."

	^(self pathFrom: anObject) reverse
! !

!MaTree methodsFor: 'testing' stamp: ''!
pathTo: anObject includes: anotherObject
	"Returns a boolean.  Example:  in the Smalltalk
	class hierarchy, pathTo: Integer includes: Magnitude would
	return true.  We use pathFrom in the implementation because
	it is faster."

	^(self pathFrom: anObject) includes: anotherObject
! !

!MaTree methodsFor: 'printing' stamp: 'cmm 12/24/2002 16:08'!
printOn: aStream

	super printOn: aStream.
	self depthFirstDo:
		[ :eachChild :eachParent |
		aStream
			cr;
			tab: (self depthOf: eachChild);
			maPrintAbbreviated: eachChild ]
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/1/2004 00:34'!
reject: aBlock 
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	^self reject: [:each | aBlock value: each ]! !

!MaTree methodsFor: 'building' stamp: 'cmm 7/25/2002 19:42'!
remove: anObject
	"Removes anObject and all of its children."

	^self
		remove: anObject
		ifAbsent: [ self error: anObject printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'building' stamp: ''!
remove: anObject ifAbsent: oneArgBlock 
	"Answers the collection of removed objects removed after removing
	anObject after removing it and all of its children."

	| tree |
	(tree := self treeAt: anObject) allMembers 
		do: [:each | parents removeKey: each]! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 6/10/2003 23:08'!
roots
	"Answers a collection of all roots in this tree."

	^self childrenOf: nil
! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 4/10/2005 18:00'!
select: aBlock 
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	| newTree |
	newTree := self maOriginalClass new.
	self allMembersDo: 
			[:each | 
			(aBlock value: each) ifTrue: [newTree addPath: (self pathTo: each)]].
	^newTree! !

!MaTree methodsFor: 'accessing' stamp: ''!
size
	^parents size! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/25/2002 19:43'!
treeAt: aNode

	"Return a new instance of the receiver which has a single main
	parent aNode."

	^self
		treeAt: aNode
		ifAbsent: [ self error: aNode printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'accessing' stamp: ''!
treeAt: anObject ifAbsent: aBlock
	"Return a new instance of the receiver which has a single main
	parent anObject."

	| newTree |

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	newTree := self class new.
	self
		depthFirstDo:
			[ :each :eachParent |
			newTree
				add: each
				asChildOf: eachParent ]
		at: anObject.
	^newTree
! !

!MaTree methodsFor: 'enumerating' stamp: 'cmm 12/24/2002 16:09'!
withParentsDo: twoArgBlock

	parents keysAndValuesDo: twoArgBlock! !
